import copy
import csv
import math
import os
from collections import defaultdict, OrderedDict
from typing import List

import matplotlib

matplotlib.use("Qt5Agg")

import matplotlib.pyplot as plt

from decide.model.base import Issue
from .. import base
from .. import calculations
from ..observers import observer


class IssueDevelopment(observer.Observer):
    """
    Keeps track of the change in issue position
    """

    def __init__(
        self,
        observable: observer.Observable,
        write_voting_position=True,
        summary_only=False,
    ):
        super().__init__(observable=observable)

        self.preference_history = {}
        self.voting_history = {}
        self.voting_loss = {}
        self.preference_loss = {}
        self.issue_obj = None
        self.write_voting_position = write_voting_position

        self.preference_history_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.voting_history_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.voting_loss_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        self.preference_loss_sum = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )

        self.denominator = 0  # TODO remove or document this attribute

        self.summary_only = summary_only

    def _setup(self):
        """
        Setup method.
        """
        for issue in self.model_ref.actor_issues:
            issue_list = {}

            for key, actor_issue in self.model_ref.actor_issues[issue].items():
                issue_list[actor_issue.actor] = []

            issue_list["nbs"] = []

            self.preference_history[issue] = copy.deepcopy(issue_list)
            self.voting_history[issue] = copy.deepcopy(issue_list)
            self.voting_loss[issue] = copy.deepcopy(issue_list)
            self.preference_loss[issue] = copy.deepcopy(issue_list)

        self.denominator += 1

    def _de_normalize_value(self, value):
        return self.issue_obj.de_normalize(value)

    def _de_normalize_list_value(self, values, _map=None):
        for _, value in enumerate(values):
            if _map is not None:
                values[_] = self.issue_obj.de_normalize(_map(value))
            else:
                values[_] = self.issue_obj.de_normalize(value)
        return values

    def _create_directories(self, repetition):
        """
        Create the directories
        :param repetition:
        :return:
        """
        if not os.path.exists(
            "{0}/issues/{1}".format(self.output_directory, repetition)
        ):
            os.makedirs("{0}/issues/{1}/csv".format(self.output_directory, repetition))
            os.makedirs(
                "{0}/issues/{1}/charts".format(self.output_directory, repetition)
            )

    def before_repetitions(self, repetitions, iterations, randomized_value=None):
        pass

    def before_iterations(self, repetition):
        """
        :param repetition:
        :return:
        """
        self._setup()

        if not self.summary_only:
            self._create_directories(repetition)

    def before_loop(self, iteration: int, repetition: int):
        pass

    def after_loop(
        self, realized: List[base.AbstractExchange], iteration: int, repetition: int
    ):
        """
        After each round we calculate the variance
        :param repetition:
        :param realized:
        :param iteration:
        :return:
        """
        model = self.model_ref

        for issue in model.actor_issues:

            nbs = model.nbs[issue]
            variance_sum = 0

            for actor_id, actor_issue in model.actor_issues[issue].items():
                # only the sum part of the variance formula
                variance_sum += (actor_issue.position - nbs) ** 2

                # the distance of the actor adjusted by his salience
                actor_loss = abs(actor_issue.position - nbs) * actor_issue.salience

                # track history
                self.preference_history[issue][actor_id].append(actor_issue.position)
                self.preference_history_sum[issue][actor_id][iteration].append(
                    actor_issue.position
                )

                # track loss
                self.preference_loss[issue][actor_id].append(actor_loss)
                self.preference_loss_sum[issue][actor_id][iteration].append(actor_loss)

            nbs_var = variance_sum / len(model.actor_issues[issue])

            self._append_nbs_preferences(
                issue=issue, iteration=iteration, nbs=nbs, nbs_var=nbs_var
            )

    def _append_nbs_preferences(self, issue, iteration, nbs, nbs_var):
        self.preference_history[issue]["nbs"].append(nbs)
        self.preference_history_sum[issue]["nbs"][iteration].append(nbs)
        self.preference_loss[issue]["nbs"].append(nbs_var)
        self.preference_loss_sum[issue]["nbs"][iteration].append(nbs_var)

    def end_loop(self, iteration: int, repetition: int):
        """
        Before each round we calculate the voting position
        :param repetition:
        :param iteration:1
        :return:
        """

        for issue in self.model_ref.actor_issues:

            nbs = self.model_ref.nbs[issue]
            variance_sum = 0

            for actor_id, actor_issue in self.model_ref.actor_issues[issue].items():
                variance_sum += (actor_issue.position - nbs) ** 2

                # the distance of the actor adjusted by his salience
                actor_loss = abs(actor_issue.position - nbs) * actor_issue.salience

                self.voting_history[issue][actor_id].append(actor_issue.position)

                self.voting_history_sum[issue][actor_id][iteration].append(
                    actor_issue.position
                )

                self.voting_loss[issue][actor_id].append(
                    abs(actor_issue.position - nbs) * actor_issue.salience
                )

                self.voting_loss_sum[issue][actor_id][iteration].append(actor_loss)

            nbs_var = variance_sum / len(self.model_ref.actor_issues[issue])

            self.voting_history[issue]["nbs"].append(nbs)
            self.voting_history_sum[issue]["nbs"][iteration].append(nbs)
            self.voting_loss[issue]["nbs"].append(nbs_var)
            self.voting_loss_sum[issue]["nbs"][iteration].append(nbs_var)

    def after_iterations(self, repetition):
        """
        Write all the data of this repetition to the filesystem
        """

        if self.summary_only:
            return

        for issue in self.preference_history:
            with open(
                "{0}/issues/{2}/csv/{1}.{3}.csv".format(
                    self.output_directory, issue, repetition, self._get_salt
                ),
                "w",
            ) as csv_file:
                writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")
                self.write_issue_after_iterations_to_file(repetition, writer, issue)

    def write_issue_after_iterations_to_file(self, repetition, writer, issue):
        self.issue_obj = self.model_ref.issues[issue]  # type: Issue

        heading = [
            "rnd-" + str(x)
            for x in range(len(self.preference_history[issue]["nbs"]))
        ]

        writer.writerow([self.issue_obj])
        writer.writerow([])  # empty row for readability

        writer.writerow(["Issue overview"])

        if self.write_voting_position:
            writer.writerow(
                [
                    "Round",
                    "Preference MDS",
                    "Preference MDS variance",
                    "",
                    "Voting MDS",
                    "Voting MDS variance",
                ]
            )
        else:
            writer.writerow(["Round", "Preference MDS", "Preference MDS variance"])

        # remove the nbs from each collection so it doesn't show up in the table.
        preference_nbs = self.preference_history[issue]["nbs"]
        del self.preference_history[issue]["nbs"]

        preference_nbs_var = self.preference_loss[issue]["nbs"]
        del self.preference_loss[issue]["nbs"]

        if self.write_voting_position:
            voting_nbs = self.voting_history[issue]["nbs"]
            del self.voting_history[issue]["nbs"]

            voting_nbs_var = self.voting_loss[issue]["nbs"]
            del self.voting_loss[issue]["nbs"]

        # create a summary table for only the preference and voting nbs + variance
        for x in range(len(preference_nbs)):

            row = [
                "rnd-" + str(x),  # round
                self._de_normalize_value(preference_nbs[x]),  # preference nbs
                self._de_normalize_value(
                    preference_nbs_var[x]
                ),  # preference nbs variance
            ]

            # write the voting positions only when the author asks for it te keep the output simple.
            if self.write_voting_position:
                row.append("")  # extra spacing for readability
                row.append(
                    self._de_normalize_value(voting_nbs[x])
                )  # voting nbs
                row.append(
                    self._de_normalize_value(voting_nbs_var[x])
                )  # voting nbs variance

            writer.writerow(row)

        # issue development summary
        writer.writerow([])
        writer.writerow(
            ["First and last round comparison of MDS and all actors"]
        )
        writer.writerow(
            [
                "Actor",
                "Salience",
                "Power",
                "First round",
                "Final round",
                "Position shift",
                "Distance MDS start",
                "Distance MDS end",
            ]
        )

        nbs_start = self._de_normalize_value(preference_nbs[0])
        nbs_end = self._de_normalize_value(preference_nbs[-1])

        writer.writerow(["MDS", "-", "-", nbs_start, nbs_end, "-"])

        # to compare different issues and variances of the model, keep the output sorted
        od = OrderedDict(sorted(self.preference_history[issue].items()))

        for actor_id, value in od.items():
            actor_issue = self.model_ref.actor_issues[self.issue_obj][actor_id]

            position_start = self._de_normalize_value(value[0])
            position_end = self._de_normalize_value(value[-1])
            position_delta = position_end - position_start

            nbs_distance_start = abs(position_start - nbs_start)
            nbs_distance_end = abs(position_start - nbs_end)
            nbs_distance_delta = nbs_distance_end - nbs_distance_start

            writer.writerow(
                [
                    actor_issue.actor,
                    actor_issue.salience,
                    actor_issue.power,
                    position_start,
                    position_end,
                    position_delta,
                    nbs_distance_start,
                    nbs_distance_end,
                    nbs_distance_delta,
                ]
            )

        # second table
        writer.writerow([])
        writer.writerow(["Preference development MDS and all actors"])
        writer.writerow(["actor", "salience", "power"] + heading)

        plt.clf()

        nbs_values = self._de_normalize_list_value(preference_nbs)

        writer.writerow(["MDS", "-", "-"] + nbs_values)

        plt.plot(nbs_values, linestyle="--",  label="MDS")

        for actor_id, value in od.items():
            actor_issue = self.model_ref.actor_issues[self.issue_obj][actor_id]

            values = self._de_normalize_list_value(value)
            writer.writerow(
                [actor_id, actor_issue.salience, actor_issue.power] + values
            )

            plt.plot(values, label=self.model_ref.actors[actor_id].name)

        lgd = plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
        plt.title(self.issue_obj)
        plt.savefig(
            "{0}/issues/{2}/charts/{1}.png".format(
                self.output_directory, self.issue_obj.name, repetition
            ),
            bbox_extra_artists=(lgd,),
            bbox_inches="tight",
        )

        if self.write_voting_position:
            writer.writerow([])
            writer.writerow(["Voting development MDS and all actors"])
            writer.writerow(["Actor", "Salience", "Power"] + heading)

            writer.writerow(
                ["MDS", "-", "-"] + self._de_normalize_list_value(voting_nbs)
            )

            od = OrderedDict(sorted(self.voting_history[issue].items()))

            for actor_id, value in od.items():
                actor_issue = self.model_ref.actor_issues[self.issue_obj][
                    actor_id
                ]
                writer.writerow(
                    [actor_id, actor_issue.salience, actor_issue.power]
                    + self._de_normalize_list_value(value)
                )

        writer.writerow([])
        writer.writerow(["Preference variance and loss of all actors"])

        writer.writerow(["nbs-var"] + preference_nbs_var)

        od = OrderedDict(sorted(self.preference_loss[issue].items()))

        for actor_id, value in od.items():
            writer.writerow([actor_id] + value)

        if self.write_voting_position:

            writer.writerow([])
            writer.writerow(["Voting variance and loss of all actors"])

            writer.writerow(["nbs-var"] + voting_nbs_var)

            od = OrderedDict(sorted(self.voting_loss[issue].items()))

            for actor_id, value in od.items():
                writer.writerow([actor_id] + value)

    def after_repetitions(self):
        """
        Write all the results to a summary with the averages for all repetitions
        :return:
        """

        self._create_directories("summary")

        for issue in self.preference_history_sum:
            with open(
                "{0}/issues/{2}/csv/{1}.{3}.{2}.csv".format(
                    self.output_directory, issue, "summary", self._get_salt
                ),
                "w",
            ) as csv_file:
                writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")

                self.issue_obj = self.model_ref.issues[issue]  # type: Issue

                heading = [
                    "rnd-" + str(x)
                    for x in range(len(self.preference_history_sum[issue]["nbs"]))
                ]

                writer.writerow([self.issue_obj])
                writer.writerow([])
                writer.writerow(["Overview issue"])

                if self.write_voting_position:
                    writer.writerow(
                        [
                            "round",
                            "avg MDS",
                            "variance",
                            "",
                            "avg voting MDS",
                            "variance",
                            "",
                            "avg MDS var",
                            "variance",
                            "",
                            "avg voting variance",
                            "variance",
                        ]
                    )
                else:
                    writer.writerow(
                        [
                            "round",
                            "avg MDS",
                            "variance",
                            "",
                            "avg voting MDS",
                            "variance",
                        ]
                    )

                preference_nbs = self.preference_history_sum[issue]["nbs"]
                del self.preference_history_sum[issue]["nbs"]

                voting_nbs = self.voting_history_sum[issue]["nbs"]
                del self.voting_history_sum[issue]["nbs"]

                preference_nbs_var = self.preference_loss_sum[issue]["nbs"]
                del self.preference_loss_sum[issue]["nbs"]

                voting_nbs_var = self.voting_loss_sum[issue]["nbs"]
                del self.voting_loss_sum[issue]["nbs"]

                _ = self._de_normalize_value

                # MDS related data.
                plt.clf()

                p_line = []

                nbs_start = None
                nbs_end = None

                nbs_voting_start = None
                nbs_voting_end = None

                for x in range(len(preference_nbs)):

                    if self.write_voting_position:

                        p = calculations.average_and_variance(preference_nbs[x])
                        v = calculations.average_and_variance(voting_nbs[x])
                        pvar = calculations.average_and_variance(preference_nbs_var[x])
                        vvar = calculations.average_and_variance(voting_nbs_var[x])

                        writer.writerow(
                            [
                                "rn-" + str(x),
                                _(p[0]),
                                p[1],
                                "",
                                _(v[0]),
                                v[1],
                                "",
                                _(pvar[0]),
                                pvar[1],
                                "",
                                _(vvar[0]),
                                vvar[1],
                            ]
                        )

                        p_line.append(_(p[0]))

                        if nbs_voting_start is None:
                            nbs_voting_start = _(v[0])
                        nbs_voting_end = _(v[0])

                    else:
                        p = calculations.average_and_variance(preference_nbs[x])
                        pvar = calculations.average_and_variance(preference_nbs_var[x])

                        writer.writerow(
                            ["rn-" + str(x), _(p[0]), p[1], "", _(pvar[0]), pvar[1]]
                        )

                        p_line.append(_(p[0]))

                    if nbs_start is None:
                        nbs_start = _(p[0])
                    nbs_end = _(p[0])

                plt.plot(p_line, linestyle="--", label="MDS")

                writer.writerow([])
                writer.writerow(
                    ["[Preference] First and last round comparison of MDS and all actors"]
                )
                writer.writerow(
                    [
                        "Actor",
                        "Salience",
                        "Power",
                        "First round",
                        "Final round",
                        "Position shift",
                        "Distance MDS start",
                        "Distance MDS end",
                    ]
                )

                writer.writerow(["MDS", "-", "-", nbs_start, nbs_end, "-"])

                for actor, value in sorted(self.preference_history_sum[issue].items()):

                    actor_issue = self.model_ref.actor_issues[issue][actor]

                    initial, var = calculations.average_and_variance(list(value.values())[0])
                    last, var = calculations.average_and_variance(list(value.values())[-1])

                    position_start = _(initial)
                    position_end = _(last)
                    position_delta = position_end - position_start

                    nbs_distance_start = abs(position_start - nbs_start)
                    nbs_distance_end = abs(position_start - nbs_end)

                    writer.writerow(
                        [
                            actor_issue.actor,
                            actor_issue.salience,
                            actor_issue.power,
                            position_start,
                            position_end,
                            position_delta,
                            nbs_distance_start,
                            nbs_distance_end,
                        ]
                    )

                writer.writerow([])
                writer.writerow(
                    ["[Voting] First and last round comparison of MDS and all actors"]
                )
                writer.writerow(
                    [
                        "Actor",
                        "Salience",
                        "Power",
                        "First round",
                        "Final round",
                        "Position shift",
                        "Distance MDS start",
                        "Distance MDS end",
                    ]
                )

                writer.writerow(["MDS", "-", "-", nbs_voting_start, nbs_voting_end, "-"])

                for actor, value in sorted(self.voting_history_sum[issue].items()):

                    actor_issue = self.model_ref.actor_issues[issue][actor]

                    initial, var = calculations.average_and_variance(list(value.values())[0])
                    last, var = calculations.average_and_variance(list(value.values())[-1])

                    position_start = _(initial)
                    position_end = _(last)
                    position_delta = position_end - position_start

                    nbs_distance_start = abs(position_start - nbs_voting_start)
                    nbs_distance_end = abs(position_start - nbs_voting_end)

                    writer.writerow(
                        [
                            actor_issue.actor,
                            actor_issue.salience,
                            actor_issue.power,
                            position_start,
                            position_end,
                            position_delta,
                            nbs_distance_start,
                            nbs_distance_end,
                        ]
                    )

                writer.writerow([])
                writer.writerow(["AVG Preference development MDS and all actors"])
                writer.writerow(["actor", "salience", "power"] + heading)

                self._write_history_sum(writer, heading, issue, self.preference_history_sum, "preference")

                writer.writerow([])
                writer.writerow(["AVG voting development MDS and all actors"])
                writer.writerow(["actor", "salience", "power"] + heading)

                if self.write_voting_position:
                    self._write_history_sum(writer, heading, issue, self.voting_history_sum, "voting")


    def _write_history_sum(self, writer, heading, issue, data, label):

        plt.clf()

        var_rows = []

        for actor, value in sorted(data[issue].items()):

            actor_issue = self.model_ref.actor_issues[issue][actor]

            row = [actor, actor_issue.salience, actor_issue.power]
            var_row = [actor, actor_issue.salience, actor_issue.power]

            avg_row = []
            for iteration, values in value.items():
                avg, var = calculations.average_and_variance(values)
                avg_row.append(self._de_normalize_value(avg))
                var_row.append(math.sqrt(var))

            plt.plot(avg_row, label=actor.name)

            writer.writerow(row + avg_row)
            var_rows.append(var_row)

        writer.writerow([])
        writer.writerow(["Standard deviation {0}".format(label)])
        writer.writerow(["actor", "salience", "power"] + heading)

        for row in var_rows:
            writer.writerow(row)

        lgd = plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
        plt.title(self.issue_obj)
        plt.savefig(
            "{0}/issues/{2}/charts/{1}.{3}.png".format(
                self.output_directory, self.issue_obj.issue_id, "summary", label
            ),
            bbox_extra_artists=(lgd,),
            bbox_inches="tight",
        )

    @property
    def _get_salt(self):
        model_name = "random"
        from decide.model.equalgain import EqualGainModel

        if isinstance(self.model_ref, EqualGainModel):
            model_name = "equal"
            if self.model_ref.randomized_value is not None:
                model_name += "-" + str(round(self.model_ref.randomized_value, 2))

        return model_name



import copy
import decimal

from decide.model import base


def nash_bargaining_solution(actor_issues, denominator):
    """
    Calculate the Nash bargaining solution.

    This is calculate in the following way:

    :math:`O_d = \\frac{\\sum_{i=1}^n C_{id} S_{id} X_{id} }{\\sum_{i=1}^n C_{id} S_{id} }`

    :param actor_issues: all the other actors on this issue
    :param denominator: the pre-calculated (and cached) denominator: :math:`\\sum_{i=1}^n C_{id} S_{id}`
    :return: Decimal
    """

    if denominator == 0:
        return 0

    numerator = 0

    for k, v in actor_issues.items():
        numerator += v.position * v.salience * v.power

    return numerator / denominator


def nbs_variance(actor_issues, nbs):
    """
    Calculate the distance between the position of the actor and the current Nash bargaining solution

    :math:`O_var = \\frac{\\sum_{i=1}^n C_{id} S_{id} X_{id} }{\\sum_{i=1}^n C_{id} S_{id} }`

    :param actor_issues: all the actors on this issue
    :param nbs: the current value of the Nash bargaining solution
    :return: the variance of positions
    """

    t = 1 / len(actor_issues)

    t2 = sum([(ai.position - nbs) ** 2 for ai in actor_issues])

    return t * t2


def calc_nbs_denominator(actor_issues):
    """
    Calculate the denominator for this issue

    :math:`\\sum_{i=1}^n C_{id} S_{id}`

    :param actor_issues: all the other actors on this issue
    :return: Decimal
    """
    denominator = 0
    for k, v in actor_issues.items():
        denominator += v.salience * v.power

    return denominator


def adjusted_nbs(actor_issues, updates, actor, new_position, denominator):
    """
    Adjust the list over actor_issues and calculates the new nash bargaining solution

    :param actor_issues: List[ActorIssue]
    :param updates: dictionary with key (actor) and value (position)
    :param actor: string, current actor
    :param new_position: int, the new position
    :param denominator: Decimal, the cached denominator
    :return: Decimal, the new nash bargaining solution
    """
    copy_ai = {}

    for k, v in actor_issues.items():
        copy_ai[v.actor] = base.ActorIssue(
            v.actor, v.issue, position=v.position, power=v.power, salience=v.salience
        )

    for key, value in updates.items():
        if key in copy_ai:  # TODO: this should not be possible
            copy_ai[key].position = value

    copy_ai[actor].position = new_position

    return nash_bargaining_solution(copy_ai, denominator)


def adjusted_nbs_by_position(actor_issues, updates, actor, x_pos, new_nbs, denominator):
    """
    Calculate the new position (delta) of the given actor when the MDS is adjusted
    :param actor_issues:
    :param updates:
    :param actor:
    :param x_pos:
    :param new_nbs:
    :param denominator:
    :return:
    """

    copy_ai = copy.deepcopy(actor_issues)

    # to be calculate:
    copy_ai[actor].position = x_pos

    for key, value in updates.items():
        copy_ai[key].position = value

    nominator = 0

    for key, value in copy_ai.items():
        if key is not actor:
            nominator += value.position * value.salience * value.power

    left = (new_nbs * denominator) - nominator

    right = left / (copy_ai[actor].salience * copy_ai[actor].power)

    return right


def reverse_move(actor_issues, actor: base.AbstractExchangeActor, exchange_ratio):
    """
    The move/shift is unknown and calculated by this method.
    It results the absolute delta/move.

    exchange_ratio * SUM[c * s] /(c * s)

    :param actor_issues:
    :param actor:
    :param exchange_ratio:
    :return:
    """
    return (exchange_ratio * sum_salience_power(actor_issues)) / (
        actor.supply.power * actor.supply.salience
    )


def by_exchange_ratio(supply_actor: base.AbstractExchangeActor, supply_exchange_ratio):
    """
    Calculate the position by exchange ratio
    TODO: better function name
    TODO: testing
    :param supply_actor: ExchangeActor
    :param supply_exchange_ratio: Decimal
    :return: Decimal
    """

    d_actor = supply_actor.opposite_actor

    sip = supply_actor.demand.salience
    sjp = d_actor.supply.salience

    sj_supply = d_actor.demand.salience
    si_supply = supply_actor.supply.salience

    return ((si_supply + sj_supply) / (sip + sjp)) * supply_exchange_ratio


def by_absolute_move(actor_issues, s_actor: base.AbstractExchangeActor):
    """

    :param actor_issues:
    :param s_actor:
    :return:
    """
    d_actor = s_actor.opposite_actor

    xip = d_actor.demand.position
    xjp = s_actor.supply.position

    sjp = s_actor.supply.salience
    cjp = s_actor.supply.power

    sum_sc = sum_salience_power(actor_issues)

    dp = (abs(xip - xjp) * sjp * cjp) / sum_sc

    return dp


def exchange_ratio(delta_x, salience, power, denominator):
    """
    Calculate the exchange ratio

    :param delta_x: the absolute distance between the old position and the new position
    :param salience: the actor salience
    :param power:  the actor power
    :param denominator: sum(c*s) for all actors on this issue
    :return: the exchange ratio
    """
    return (delta_x * salience * power) / denominator


def sum_salience_power(actor_issues):
    """
    Helper method to calculate the sum of the power * salience of all the ActorIssues
    :param actor_issues:
    :return:
    """
    return sum(actor.power * actor.salience for actor in actor_issues.values())


def expected_utility(actor, demand_exchange_ratio, supply_exchange_ratio):
    """

    :param actor:
    :param demand_exchange_ratio:
    :param supply_exchange_ratio:
    :return:
    """
    return abs(
        demand_exchange_ratio * actor.supply.salience
        - supply_exchange_ratio * actor.demand.salience
    )


def is_gain_equal(eui, euj, threshold=1e-20):
    """

    :param eui: The utility gain of actor i
    :param euj: The utility gain of actor j
    :param threshold: The threshold where the diff needs to be below
    :return: True when the gains are (approximately) equal
    """
    if abs(eui - euj) > threshold:
        raise Exception(
            "Expected equal gain but found {0} and {1} results in {2}. "
            "Adjust the above threshold to a higher value and continue.".format(
                eui, euj, abs(eui - euj)
            )
        )

    return True


def actor_externalities(
    actor: base.Actor, model_ref: base.AbstractModel, realized: base.AbstractExchange
):
    """
    Calculate the externalities from an exchange

    :param actor: current actor
    :param model_ref: model
    :param realized: realized exchanges
    :return: the Decimal value of the externality
    """

    if (
        actor in model_ref.actor_issues[realized.j.supply.issue]
        and actor in model_ref.actor_issues[realized.i.supply.issue]
    ):

        xp = model_ref.actor_issues[realized.j.supply.issue][actor].position
        sp = model_ref.actor_issues[realized.j.supply.issue][actor].salience

        xq = model_ref.actor_issues[realized.i.supply.issue][actor].position
        sq = model_ref.actor_issues[realized.i.supply.issue][actor].salience

        l0 = abs(realized.j.nbs_0 - xp)
        l1 = abs(realized.j.nbs_1 - xp)
        r0 = abs(realized.i.nbs_0 - xq)
        r1 = abs(realized.i.nbs_1 - xq)

        l = l0 - l1
        r = r0 - r1
        ext = l * sp + r * sq

        return ext
    return 0


def position_by_nbs(actor_issues, exchange_actor, nbs, denominator):
    """
    For the Random Rate implementation the position is need to be calculated where a MDS is given.

    :param actor_issues:
    :param exchange_actor:
    :param nbs:
    :param denominator:
    :return:
    """
    # TODO: add the math formula to the docstring

    sum_c_s_x = 0

    for actor_name, actor_issue in actor_issues.items():
        if actor_name != exchange_actor.actor_name:
            sum_c_s_x += actor_issue.salience * actor_issue.power * actor_issue.position

    return (nbs * denominator - sum_c_s_x) / (exchange_actor.c * exchange_actor.s)


def average_and_variance(values: list):
    count = len(values)

    if count == 0:
        return 0, 0

    average = sum(values) / count

    variance = sum([(x - average) ** 2 for x in values]) / count

    return average, variance


def exchange_ratio_by_expected_utility(delta_q, sq, sp, utility=0):
    """
    Calculates the exchange ratio by a gain of 0.
    Used for the new model
    :param utility: Always 0
    :param delta_q:  Exchange Ratio
    :param sq: Salience
    :param sp: Salience
    :return:
    """
    return (utility + (delta_q * sq)) / sp


def new_start_position(salience, x, y, salience_weight=0.4, fixed_weight=0.1):
    sw = decimal.Decimal(salience_weight)
    fw = decimal.Decimal(fixed_weight)
    swv = (1 - salience) * sw * y
    fwv = fw * y
    pv = (1 - (1 - salience) * sw - fw) * x

    return swv + fwv + pv



import csv
from decimal import Decimal
from pathlib import Path

import matplotlib.pyplot as plt


def append_nbs(data):

    saliences = {}
    powers = {}
    positions = {}
    actors = []

    denominator = Decimal(0)
    power_salience_positions = [Decimal(0) for _ in range(10)]

    for actor, salience, power, *_ in data:

        salience = Decimal(salience)
        power = Decimal(power)

        saliences[actor] = salience
        powers[actor] = power
        positions[actor] = _
        actors.append(actor)

        power_salience = power * salience

        denominator += power_salience

        for index, position in enumerate(_):
            power_salience_positions[index] += power_salience * Decimal(position)

    return [x / denominator for x in power_salience_positions]


def graph_builder(file_path: str, p: str):
    path = Path(file_path)

    csv_path = path / "csv"
    chart_path = path / "charts" / "modified"

    # chart_path_2 = Path("/home/jelmert/Downloads/kopenhagen (1)/kopenhagen/modified")

    if not chart_path.exists():
        chart_path.mkdir()

    for file in csv_path.glob("*.csv"):
        with file.open() as csv_file:
            reader = csv.reader(csv_file, delimiter=";", quotechar="'")

            plot_name = next(reader)[0] + f" p={p}"

            chart_area = get_chart_area_from_csv(
                "AVG Preference development MDS and all actors", reader
            )

            nbs = append_nbs(chart_area)

            plt.plot(nbs, linestyle="--", label="MDS")

            plt.title(plot_name)

            for row in chart_area:
                plt.plot([float(x) for x in row[3:]], label=row[0])

            x = chart_path / file.name
            lgd = plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
            plt.savefig(
                str(x).replace(".csv", ".png"),
                bbox_extra_artists=(lgd,),
                bbox_inches="tight",
            )
            plt.show()
            plt.clf()


def get_chart_area_from_csv(begin: str, reader: csv.reader):
    data = []

    found_begin = False

    for index, row in enumerate(reader):

        if len(row) > 0 and row[0] == begin:
            found_begin = index

        if found_begin:
            if len(row) == 0 or row[0] == "":
                return data

            if len(row) > 0 and index - found_begin > 1:
                data.append(row)

    return data


if __name__ == "__main__":

    for p in ["0.00", "0.20", "0.40", "0.60", "0.80", "1.00"]:
        graph_builder(
            # ~Alternative power files artikel/Alternative power files artikel/CoP21_2Oct2015_A_no_allfin_amb alt power/CoP21_2Oct2015_A_no_allfin_amb alt power/1.00/issues/summary
            f"/home/jelmert/Downloads/Alternative power files artikel/Alternative power files artikel/CoP21_2Oct2015_A_no_allfin_amb alt power/CoP21_2Oct2015_A_no_allfin_amb alt power/{p}/issues/summary",
            p
        )



from unittest import TestCase

from decide.model.equalgain import EqualGainModel, EqualGainExchange


class TestExchange(TestCase):
    def test_init(self):
        model = EqualGainModel()

        a1 = model.add_actor("a1")
        a2 = model.add_actor("a2")
        a3 = model.add_actor("a3")
        a4 = model.add_actor("a4")

        p = model.add_issue("p")
        p.lower = 0
        p.upper = 100

        q = model.add_issue("q")
        q.lower = 0
        q.upper = 100

        model.add_actor_issue(a1, "p", 70, 0.75, 0.5)
        model.add_actor_issue(a2, "p", 30, 0.25, 0.5)
        model.add_actor_issue(a3, "p", 30, 0.25, 0.5)
        model.add_actor_issue(a4, "p", 0, 0.6, 0.85)

        model.add_actor_issue(a1, "q", 70, 0.25, 0.5)
        model.add_actor_issue(a2, "q", 30, 0.75, 0.5)
        model.add_actor_issue(a3, "q", 30, 0.75, 0.5)
        model.add_actor_issue(a4, "q", 100, 0.8, 0.85)

        model.calc_nbs()

        e = EqualGainExchange(a1, a2, "p", "q", model, groups=["a", "d"])
        e.calculate()

        e2 = EqualGainExchange(a1, a2, "p", "q", model, groups=["a", "d"])

        e2.updates["p"]["a3"] = 100
        e2.calculate()

        e3 = EqualGainExchange(a1, a3, "p", "q", model, groups=["a", "d"])
        e3.calculate()



import os

import pandas as pd
from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)


def write_summary_result(conn, model_run_ids, output_directory):
    df = pd.read_sql("""
    SELECT
            a.name AS actor,
            i.name as issue,
            AVG(ai.position) as position,        
            i2.pointer + 1                                AS round,
            m.p,
            m.id  
          FROM actorissue ai
            LEFT JOIN issue i ON ai.issue_id = i.id
            LEFT JOIN actor a ON ai.actor_id = a.id
            LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
            LEFT JOIN repetition r ON i2.repetition_id = r.id
            LEFT JOIN modelrun m ON r.model_run_id = m.id            
          WHERE  ai.type = 'before' AND m.id IN(%s)
         GROUP BY m.id, i2.pointer, a.id, i.id;
    """ % list_to_sql_param(model_run_ids),
                     conn,
                     index_col=['issue', 'actor', 'p'],
                     columns=['position']
                     )

    table = pd.pivot_table(df, index=['issue', 'actor', 'round'], columns=['p'], values=['position'])
    table.to_csv(os.path.join(output_directory, 'issues_preference.csv'))

    df = pd.read_sql("""
    SELECT
            a.name AS actor,
            i.name as issue,
            AVG(ai.position) as position,        
            i2.pointer + 1                                AS round,
            m.p,
            m.id  
          FROM actorissue ai
            LEFT JOIN issue i ON ai.issue_id = i.id
            LEFT JOIN actor a ON ai.actor_id = a.id
            LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
            LEFT JOIN repetition r ON i2.repetition_id = r.id
            LEFT JOIN modelrun m ON r.model_run_id = m.id            
          WHERE  ai.type = 'after' AND m.id IN(%s)
         GROUP BY m.id, i2.pointer, a.id, i.id;
    """ % list_to_sql_param(model_run_ids),
                     conn,
                     index_col=['issue', 'actor', 'p'],
                     columns=['position']
                     )

    table = pd.pivot_table(df, index=['issue', 'actor', 'round'], columns=['p'], values=['position'])
    table.to_csv(os.path.join(output_directory, 'issues_voting.csv'))


if __name__ == '__main__':
    m = Manager(os.environ.get('DATABASE_URL'))
    m.init_database()

    model_run_id = 1

    write_summary_result(connection, [model_run_id], data_folder)



import decimal
from unittest.mock import MagicMock, patch

import pytest

from decide.model import calculations, base
from decide.model.equalgain import EqualGainModel, EqualGainExchange


class TestModel:
    """
    Object tit populate with test data
    """

@pytest.fixture
def test_model():
    test = TestModel()

    a = base.Actor("a")
    b = base.Actor("b")
    c = base.Actor("c")

    p = base.Issue("p")

    test.a = base.ActorIssue(a, p, position=0, salience=1, power=1)
    test.b = base.ActorIssue(b, p, position=100, salience=1, power=1)
    test.c = base.ActorIssue(c, p, position=100, salience=1, power=1)

    test.actor_issues = {"a": test.a, "b": test.b, "c": test.c}
    test.denominator = calculations.calc_nbs_denominator(test.actor_issues)
    test.model = EqualGainModel()

    return test


def test_calc_nbs(test_model):
    denominator = calculations.calc_nbs_denominator({"a": test_model.a, "b": test_model.b})

    assert calculations.nash_bargaining_solution({"a": test_model.a, "b": test_model.b}, denominator) == 50
    assert calculations.nash_bargaining_solution(test_model.actor_issues, test_model.denominator) == decimal.Decimal(
        200) / decimal.Decimal(3)


def test_calc_adjusted_nbs(test_model):
    a1 = "a"
    a2 = "b"
    a3 = "c"

    assert calculations.adjusted_nbs(test_model.actor_issues, {}, a1, 100, test_model.denominator) == 100

    assert calculations.adjusted_nbs(test_model.actor_issues, {}, a2, 0, test_model.denominator) == decimal.Decimal(
        100) / decimal.Decimal(3)
    assert calculations.adjusted_nbs(test_model.actor_issues, {}, a1, 0, test_model.denominator) == decimal.Decimal(
        200) / decimal.Decimal(3)
    assert calculations.adjusted_nbs(test_model.actor_issues, {"a": 100}, a3, 0,
                                     test_model.denominator) == decimal.Decimal(200) / decimal.Decimal(3)


def test_by_absolute_move(sample_model):
    model = sample_model

    i = model.actors["Wcentr"]
    j = model.actors["Willems"]
    p = model.issues["tolheffing-binnenstad"]
    q = model.issues["extra-investering"]

    e = EqualGainExchange(i, j, p, q, model, groups=["a", "d"])

    dp = calculations.by_absolute_move(model.actor_issues[e.i.supply.issue], e.i)
    dq = calculations.by_exchange_ratio(e.i, dp)

    assert dp - decimal.Decimal(13.15789473684210526315789474) < 0.001
    assert dq - decimal.Decimal(10.38781163434903047091412743) < 0.001

    move = calculations.reverse_move(
        model.actor_issues[e.j.supply.issue], exchange_ratio=dq, actor=e.j
    )

    assert move - decimal.Decimal(61.05117364047237206589881911) < 0.001

    assert move < abs(e.i.supply.position - e.j.demand.position)

    dp_1 = calculations.by_absolute_move(model.actor_issues[e.j.supply.issue], e.j)
    dq_1 = calculations.by_exchange_ratio(e.j, dp)

    move_1 = calculations.reverse_move(
        model.actor_issues[e.i.supply.issue], e.i, dq_1
    )

    assert move_1 > abs(e.i.supply.position - e.j.demand.position)


def test_sum_salience_power(test_model):
    assert calculations.sum_salience_power({"a": test_model.a, "b": test_model.b, "c": test_model.c}) == 3


def test_new_start_position():
    assert calculations.new_start_position(1, 0, 100) - 10 < 1e-10
    assert calculations.new_start_position(1, 100, 0) - 90 < 1e-10


@patch("decide.model.calculations.sum_salience_power")
def test_reverse_move(sum_salience_power):
    sum_salience_power.return_value = 3
    actor = MagicMock()
    actor.supply = MagicMock(salience=1, power=1)

    res = calculations.reverse_move(actor_issues=[], actor=actor, exchange_ratio=1)

    assert res == 3


def test_sum_salience_power_2():
    actor_issues = [
        base.ActorIssue(None, None, 1, 1, 1),
        base.ActorIssue(None, None, 1, 1, 1),
    ]

    mock = MagicMock()
    mock.values = MagicMock(return_value=actor_issues)

    assert mock.values(), actor_issues

    assert calculations.sum_salience_power(mock), 2

    actor_issues.append(base.ActorIssue(None, None, 1, 1, 1))

    assert calculations.sum_salience_power(mock) == 3



from unittest import TestCase

from decide.model.base import AbstractExchangeActor, AbstractModel, Issue


class TestAbstractExchangeActor(TestCase):
    def test___eq__(self):
        model = AbstractModel()

        test_issue = model.add_issue("test")
        test1_issue = model.add_issue("test1")

        test_issue.lower = 0
        test_issue.upper = 100
        test_issue.calculate_step_size()

        test1_issue.lower = 0
        test1_issue.upper = 100
        test1_issue.calculate_step_size()

        test_actor = model.add_actor("test")
        test1_actor = model.add_actor("test1")
        test1_issue_duplicated = model.add_issue("test")
        test1_actor_duplicated = model.add_actor("test1")

        test1_issue_duplicated.lower = 0
        test1_issue_duplicated.upper = 100

        self.assertEqual(test_actor, test_actor)
        self.assertNotEqual(test_actor, test1_actor)

        self.assertEqual(test_actor.__hash__(), hash(test_actor.actor_id))
        self.assertEqual(test1_actor_duplicated, test1_actor)

        self.assertTrue(test_actor in model.actors)
        self.assertTrue(test_issue in model.issues)

        b = Issue("test")
        b.name = "test"
        b.lower = 0
        b.upper = 100
        b.calculate_step_size()

        self.assertEqual(b.__hash__(), test_issue.__hash__())

        self.assertTrue(b in model.issues)

        self.assertEqual(test1_issue_duplicated, test_issue)
        self.assertEqual(test_issue.__hash__(), hash(test_issue.issue_id))
        self.assertEqual(test_issue.__hash__(), hash("test"))
        self.assertEqual(test_issue.__hash__(), hash(test1_issue_duplicated.issue_id))

        self.assertTrue(test_actor in model.actors)
        ai1 = model.add_actor_issue("test", "test", 100, 1, 1)
        ai2 = model.add_actor_issue("test", "test1", 100, 1, 1)
        ai3 = model.add_actor_issue("test1", "test", 100, 1, 1)
        ai4 = model.add_actor_issue("test1", "test1", 100, 1, 1)

        exchange_actor0 = AbstractExchangeActor(
            model=model,
            actor=test_actor,
            demand_issue=test_issue,
            supply_issue=test1_issue,
            exchange=None,
        )

        exchange_actor1 = AbstractExchangeActor(
            model=model,
            actor=test1_actor,
            demand_issue=test1_issue,
            supply_issue=test_issue,
            exchange=None,
        )

        exchange_actor4 = AbstractExchangeActor(
            model=model,
            actor=test_actor,
            demand_issue=test_issue,
            supply_issue=test1_issue,
            exchange=None,
        )

        self.assertTrue(exchange_actor0 == exchange_actor0)
        self.assertTrue(exchange_actor0 == exchange_actor4)
        self.assertTrue(exchange_actor0 != exchange_actor1)

        self.assertTrue(exchange_actor0.equals_supply_issue(exchange_actor4))
        self.assertTrue(exchange_actor0.equals_demand_issue(exchange_actor4))

        self.assertFalse(exchange_actor0.equals_supply_issue(exchange_actor1))
        self.assertFalse(exchange_actor4.equals_supply_issue(exchange_actor1))



import logging
import os
import sys

import requests
from PyQt5 import QtWidgets, QtGui

from decide import log_filename, input_folder
from decide.data.reader import InputDataFile
from decide.qt.inputwindow import signals
from decide.qt.inputwindow.models import IssueInputModel, ActorInputModel
from decide.qt.inputwindow.widgets import (
    ActorWidget,
    IssueWidget,
    ActorIssueWidget,
    PositionWidget,
    SalienceWidget,
)
from decide.qt.mainwindow.gui import DecideMainWindow
from decide.qt.mainwindow.helpers import esc, normalize
from decide.qt.utils import show_user_error


class InputWindow(QtWidgets.QMainWindow):
    def __init__(self, main_window: DecideMainWindow = None, *args, **kwargs):
        super(InputWindow, self).__init__(*args, **kwargs)

        # save the filename which we are editing
        self.input_filename = None

        # refresh the main windows after editing
        self.main_window = main_window

        self.main = QtWidgets.QHBoxLayout()
        self.left = QtWidgets.QVBoxLayout()

        self.tabs = QtWidgets.QTabWidget()

        self.actor_widget = ActorWidget()
        self.issue_widget = IssueWidget()

        self.actor_issue_widget = ActorIssueWidget(self.actor_widget, self.issue_widget)

        self.position_widget = PositionWidget(self.actor_issue_widget)
        self.salience_widget = SalienceWidget(self.actor_issue_widget)

        self.left.addWidget(self.actor_widget)
        self.left.addWidget(self.issue_widget)

        self.tabs.addTab(self.position_widget, "Positions (by Issue)")
        self.tabs.addTab(self.salience_widget, "Salience (by Actor)")

        self.main.addLayout(self.left)
        self.main.addWidget(self.tabs)

        q = QtWidgets.QWidget()
        q.setLayout(self.main)

        self.menubar = QtWidgets.QMenuBar(self)

        self.init_menu()
        self.setCentralWidget(q)

        self.setGeometry(300, 300, 1024, 768)
        self.setWindowTitle("Decide Exchange Model")
        self.show()

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        if self.input_filename:
            self.main_window.load_input_data(self.input_filename)

    def init_menu(self):

        self.menubar.setNativeMenuBar(False)
        self.setMenuBar(self.menubar)

        file_menu = self.menubar.addMenu("File")
        example_menu = self.menubar.addMenu("Examples")

        load_kopenhagen = QtWidgets.QAction("&load Kopenhagen", self.menubar)
        load_kopenhagen.triggered.connect(self.load_copenhagen)

        load_cop = QtWidgets.QAction("&load Parijs", self.menubar)
        load_cop.triggered.connect(self.load_cop21)

        open_action = QtWidgets.QAction("Open", self.menubar)
        open_action.triggered.connect(self.open_dialog)

        save_action = QtWidgets.QAction("Save", self.menubar)
        save_action.triggered.connect(self.save_location)

        example_menu.addAction(load_kopenhagen)
        example_menu.addAction(load_cop)

        file_menu.addAction(open_action)
        file_menu.addAction(save_action)

    def load_copenhagen(self):
        self.load_input_file(
            "https://raw.githubusercontent.com/foarsitter/decide-exchange-model/master/data/input/copenhagen.csv")

    def load_sample_data(self):
        self.load_input_file(
            "https://raw.githubusercontent.com/foarsitter/decide-exchange-model/master/data/input/sample_data.txt")

    def load_cop21(self):
        self.load_input_file(
            "https://raw.githubusercontent.com/foarsitter/decide-exchange-model/master/data/input/cop21.csv"
        )

    def load_input_file(self, file_path):

        if file_path.startswith('http'):
            response = requests.get(file_path)

            file_path = file_path.split('/')[-1]

            download_dir = os.path.join(input_folder, 'downloads')

            os.makedirs(download_dir, exist_ok=True)

            with open(os.path.join(download_dir, file_path), 'wb') as file:
                file.write(response.content)

        file = os.path.join(input_folder, file_path)

        if os.path.isfile(file):
            self.load(file)
        else:
            show_user_error(self, "Cannot load {}".format(file))

    def open_dialog(self):

        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Select input data")

        if file_name:
            try:
                self.load(file_name)
            except Exception as ex:
                error_dialog = QtWidgets.QErrorMessage(self)
                error_dialog.showMessage(str(ex))

                raise ex

    def save_location(self):

        if self.input_filename:

            button_reply = QtWidgets.QMessageBox.question(
                self,
                "Save",
                "Overwrite existing file {}".format(
                    self.input_filename
                ),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )
        else:
            button_reply = QtWidgets.QMessageBox.No

        if button_reply == QtWidgets.QMessageBox.Yes:
            file_name = self.input_filename
        else:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Select input data")

        if file_name:
            self.save(file_name)

    def load(self, input_filename):

        data_file = InputDataFile.open(input_filename)

        actor_inputs = {}
        issue_inputs = {}
        actor_powers = {}

        self.actor_widget.clear()
        self.issue_widget.clear()
        self.actor_issue_widget.clear()
        self.position_widget.clear()
        self.salience_widget.clear()

        # collect all the power values for the Actor widget
        for actor_issue in data_file.actor_issues.values():
            if actor_issue.actor not in actor_powers:
                actor_powers[actor_issue.actor] = actor_issue.power

        for actor in data_file.actors.values():
            actor_input = self.actor_widget.add_actor(
                actor.id, actor_powers[actor.id]
            )

            actor_input.comment = str(actor.comment)

            actor_inputs[actor] = actor_input

            self.actor_issue_widget.actors.add(actor_input)

        for issue in data_file.issues.values():
            issue_input = self.issue_widget.add_issue(
                issue.name, issue.lower, issue.upper
            )
            issue_input.comment = issue.description

            issue_inputs[issue_input.name] = issue_input

            self.actor_issue_widget.issues.add(issue_input)

        for actor_issue in data_file.actor_issues.values():
            actor_input = actor_inputs[actor_issue.actor]
            issue_input = issue_inputs[actor_issue.issue]

            self.actor_issue_widget.add_actor_issue(
                actor_input, issue_input, actor_issue
            )

        self.position_widget.redraw()
        self.position_widget.set_choices(self.actor_issue_widget.issues)

        self.salience_widget.redraw()
        self.salience_widget.set_choices(self.actor_issue_widget.actors)

        self.input_filename = input_filename

    def save(self, filename):

        with open(filename, "w") as file:

            powers = {}

            for actor in self.actor_widget.actor_issues.values():
                powers[actor.name] = actor.power

                file.write(
                    "\t".join(
                        [
                            esc("#A"),
                            esc(actor.name),
                            esc(actor.name),
                            esc(actor.comment),
                            "\n",
                        ]
                    )
                )

            for issue in self.issue_widget.actor_issues.values():
                file.write(
                    "\t".join(
                        [
                            esc("#P"),
                            esc(issue.name),
                            esc(issue.name),
                            esc(issue.comment),
                            "\n",
                        ]
                    )
                )

                file.write(
                    "\t".join([esc("#M"), esc(issue.name), str(issue.lower), "\n"])
                )

                file.write(
                    "\t".join([esc("#M"), esc(issue.name), str(issue.upper), "\n"])
                )

            for actor_id, actor_issues in self.actor_issue_widget.actor_issues.items():

                for issue_id, actor_issue in actor_issues.items():
                    actor_issue = actor_issue  # type: ActorIssueInputModel

                    file.write(
                        "\t".join(
                            [
                                esc("#D"),
                                esc(actor_issue.actor.name),
                                esc(actor_issue.issue.name),
                                normalize(actor_issue.position),
                                normalize(actor_issue.salience),
                                normalize(powers[actor_issue.actor.name]),
                                "\n",
                            ]
                        )
                    )

        # open_file_natively(filename)


input_window = None


@signals.actor_created.connect
def actor_issue_box_actor_created(sender: ActorInputModel):
    input_window.actor_issue_widget.add_actor(sender)

    input_window.salience_widget.add_choice(sender.id, sender.name)


@signals.actor_deleted.connect
def actor_issue_box_delete_actor(sender: ActorInputModel, **kwargs):
    input_window.actor_issue_widget.delete_actor(sender)
    input_window.salience_widget.delete_choice(sender.id)


@signals.actor_changed.connect
def actor_issue_box_update_actor(sender: ActorInputModel, key: str, value):
    if key == "name":
        input_window.salience_widget.update_choice(sender.id, sender.name)
    elif key == "power":

        for actor_issue in input_window.actor_issue_widget.actor_issues[sender.id].values():
            actor_issue.set_power(value, True)

    input_window.position_widget.redraw()
    input_window.salience_widget.redraw()


@signals.issue_created.connect
def actor_issue_box_issue_created(sender: IssueInputModel):
    input_window.actor_issue_widget.add_issue(sender)

    input_window.position_widget.add_choice(sender.id, sender.name)


@signals.issue_deleted.connect
def actor_issue_box_delete_issue(sender: IssueInputModel, **kwargs):
    input_window.actor_issue_widget.delete_issue(sender)
    input_window.position_widget.delete_choice(sender.id)


@signals.issue_changed.connect
def actor_issue_box_update_issue(sender: IssueInputModel, key: str, value):
    if key == "name":
        input_window.position_widget.update_choice(sender.id, sender.name)

    input_window.position_widget.redraw()
    input_window.salience_widget.redraw()


def register_app(ex):
    global input_window
    input_window = ex


if __name__ == "__main__":
    logging.basicConfig(
        filename=log_filename,
        filemode="w",
        level=logging.DEBUG,
        format=" %(asctime)s - %(levelname)s - %(message)s",
    )

    #    sys.excepthook = exception_hook

    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)

    input_window = InputWindow()

    sys.exit(app.exec_())



import logging
import os

import pandas as pd
from pandas.errors import DataError

from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param, handle_data_frame

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)


def write_summary_result(conn, model_run_ids, output_directory):
    sql = """SELECT SUM(e.own)            as own,
           SUM(e.inner_positive) as inner_positive,
           SUM(e.inner_negative) as inner_negative,
           SUM(e.outer_positive) as outer_positive,
           SUM(e.outer_negative) as outer_negative,
           a.name                   actor,
           m.p
    FROM externality e
             JOIN actor a on e.actor_id = a.id
             JOIN iteration i on e.iteration_id = i.id
             JOIN repetition r on i.repetition_id = r.id
             JOIN modelrun m on r.model_run_id = m.id         
    WHERE m.id IN(%s)
    GROUP BY e.actor_id, m.id""" % list_to_sql_param(model_run_ids)

    df = pd.read_sql_query(
        sql,
        con=conn,
        index_col=['p'],
    )

    try:
        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['inner_positive']),
            file_name=os.path.join(output_directory, 'externalities_inner_positive.{}'),
            title='Inner positive externalities',
        )
    except DataError as ex:
        logging.exception(ex)
        print(ex)
    try:
        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['inner_negative']),
            file_name=os.path.join(output_directory, 'externalities_inner_negative.{}'),
            title='Inner negative externalities',
        )
    except DataError as ex:
        logging.exception(ex)
        print(ex)
    try:

        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['outer_positive']),
            file_name=os.path.join(output_directory, 'externalities_outer_positive.{}'),
            title='Outer positive externalities',
        )

    except DataError as ex:
        logging.exception(ex)
        print(ex)
    try:

        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['outer_negative']),
            file_name=os.path.join(output_directory, 'externalities_outer_negative.{}'),
            title='Outer negative externalities',
        )

    except DataError as ex:
        logging.exception(ex)
        print(ex)
    try:

        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['own']),
            file_name=os.path.join(output_directory, 'externalities_own.{}'),
            title='Own utility sun',
        )
    except DataError as ex:
        logging.exception(ex)
        print(ex)


if __name__ == '__main__':
    m = Manager(os.environ.get('DATABASE_URL'))
    m.init_database()

    model_run_id = 1

    write_summary_result(connection, [model_run_id], data_folder)



import argparse
import logging
import os
from datetime import datetime
from typing import List

from decide import input_folder
from decide.data.modelfactory import ModelFactory
from decide.data.reader import InputDataFile
from decide.model import randomrate, equalgain
from decide.model.observers.exchanges_writer import ExchangesWriter
from decide.model.observers.externalities import Externalities
from decide.model.observers.issue_development import IssueDevelopment
from decide.model.observers.logger import Logger
from decide.model.observers.observer import Observable
from decide.model.observers.sqliteobserver import SQLiteObserver
from decide.model.utils import ModelLoop


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Decide exchange model".format()
    )
    parser.add_argument(
        "--model",
        "-m",
        help='The type of the model. The options are "equal" for the Equal Gain model and '
             '"random" for the RandomRate model ',
        default="equal",
        type=str,
    )

    parser.add_argument(
        "--p", "-p", help="Randomized Equal Gain", default=None, type=str
    )
    parser.add_argument(
        "--iterations",
        "-i",
        help="The number of round the model needs to be executed",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--repetitions",
        "-r",
        help="How many times it has te be repeated?",
        default=1,
        type=int,
    )
    parser.add_argument(
        "--input_file",
        help="The location of the csv input file. ",
        default=os.path.join(input_folder, 'sample_data.txt'),
        type=str,
    )
    parser.add_argument(
        "--output_dir", help="Output directory ", default="../data/output/", type=str
    )
    parser.add_argument(
        "--database",
        help="The SQLite database",
        default="../data/output/decide-data_1.db",
        type=str,
    )

    parser.add_argument("--step", default='0.80', type=str)
    parser.add_argument("--stop", default='0.80', type=str)
    parser.add_argument("--start", default='0.0', type=str)
    parser.add_argument("--name", default=None, type=str)
    parser.add_argument("--actors", default=None, type=str)
    parser.add_argument("--issues", default=None, type=str)

    return parser.parse_args()


def init_event_handlers(model, output_directory, database_file, write_csv=True):
    event_handler = Observable(model_ref=model, output_directory=output_directory)

    SQLiteObserver(event_handler, output_directory)

    Logger(event_handler)
    Logger.LOG_LEVEL = 99

    if write_csv:
        # csv handlers
        Externalities(event_handler)
        ExchangesWriter(event_handler)
        IssueDevelopment(event_handler)

    return event_handler


def init_output_directory(*args):
    output_directory = os.path.join(*args)

    if not os.path.isdir(output_directory):

        if os.path.isfile(output_directory):
            output_directory += '_output'

        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)

    return output_directory


def float_range(start=0.0, stop=1.0, step=0.05):
    # when the step size is 0, return
    if step == 0:
        yield 0
    else:
        # prevent locking yourself out
        if (stop - start) / step > 256:
            raise RuntimeError("Maximum steps exceeded with step={} (256)".format(step))

        i = start
        # add a 10th step to stop for floating point rounding differences 0.500001 vs 0.499999
        while i < (stop + step / 10):
            yield i
            i += step


def actors_param(args) -> List[str]:
    actors = None
    if args.actors:
        actors = args.actors.split(
            ";"
        )

    return actors


def issues_param(args) -> List[str]:
    issues = None

    if args.issues:
        issues = args.issues.split(
            ";"
        )

    return issues


def p_values_param(args) -> List[float]:
    p_values = []

    if args.start and args.step and args.stop:
        p_values = [
            str(round(p, 2))
            for p in float_range(
                start=float(args.start), step=float(args.step), stop=float(args.stop)
            )
        ]

    elif args.p:

        if ";" in args.p:
            p_values += args.p.split(";")
        else:
            p_values.append(args.p)

    return p_values


def main():
    args = parse_arguments()

    p_values = p_values_param(args)
    issues = issues_param(args)
    actors = actors_param(args)

    data_file = InputDataFile.open(args.input_file)

    data_set_name = os.path.splitext(os.path.basename(args.name or args.input_file))[0]

    factory = ModelFactory(data_file, actor_whitelist=actors, issue_whitelist=issues)

    """
        Initial the right model from the given arguments
        """

    if args.model == "equal":
        model_klass = equalgain.EqualGainModel
    else:
        model_klass = randomrate.RandomRateModel

    for p in p_values:

        start_time = datetime.now()  # for timing operations

        model = factory(model_klass=model_klass, randomized_value=p)

        output_directory = init_output_directory(
            args.output_dir,
            data_set_name,
            p,
        )

        # The event handlers for logging and writing the results to the disk.
        event_handler = init_event_handlers(
            model=model,
            output_directory=output_directory,
            database_file=args.database,
            write_csv=True,
        )

        event_handler.log(message="Start calculation at {0}".format(start_time))

        event_handler.log(message="Parsed file".format(args.input_file))

        event_handler.before_repetitions(
            repetitions=args.repetitions, iterations=args.iterations
        )

        for repetition in range(args.repetitions):

            model = factory(model_klass=model_klass, randomized_value=p)

            event_handler.update_model_ref(model)

            model_loop = ModelLoop(model, event_handler, repetition)

            event_handler.before_iterations(repetition)

            for iteration_number in range(args.iterations):
                logging.info("round {0}.{1}".format(repetition, iteration_number))
                model_loop.loop()

            event_handler.after_iterations(repetition)

        event_handler.after_repetitions()

        event_handler.log(message="Finished in {0}".format(datetime.now() - start_time))

    logging.info("Done")


if __name__ == "__main__":
    main()



from PyQt5 import QtCore

from decide.qt.inputwindow.gui import InputWindow, register_app


def test_actor_widget_add_button(qtbot):
    x = InputWindow()
    register_app(x)

    assert x.actor_widget._row_pointer == 1, 'Only the heading is shown'
    assert x.actor_widget.grid_layout.rowCount() == 1

    qtbot.mouseClick(x.actor_widget.add_button, QtCore.Qt.LeftButton)

    assert x.actor_widget.grid_layout.rowCount() == 2, 'The heading and an empty row is shown'



import csv
import decimal
import random
import statistics
from decimal import Decimal

import pytest

from decide.data.modelfactory import ModelFactory
from decide.data.reader import InputDataFile
from decide.model.equalgain import EqualGainModel, EqualGainExchange, EqualGainExchangeActor


@pytest.fixture
def model_factory():
    input_data = """'#A'	'Actor-1'	'Actor-1'	''	
'#A'	'Actor-2'	'Actor-2'	''	
'#P'	'Issue-1'	'Issue-1'	'Issue-1'	
'#M'	'Issue-1'	0	
'#M'	'Issue-1'	100	
'#P'	'Issue-2'	'Issue-2'	'Issue-2'	
'#M'	'Issue-2'	0	
'#M'	'Issue-2'	100	
'#D'	'Actor-1'	'Issue-1'	0.0	10	1	
'#D'	'Actor-1'	'Issue-2'	0.0	90	1	
'#D'	'Actor-2'	'Issue-1'	100.0	60	1	
'#D'	'Actor-2'	'Issue-2'	100.0	50	1"""

    reader = csv.reader(input_data.splitlines(), delimiter="\t", quotechar="'")

    data_file = InputDataFile.open_reader(reader)

    assert len(data_file.actors) == 2
    assert len(data_file.actor_issues) == 4

    model_factory = ModelFactory(data_file)

    return model_factory


def test_rex(model_factory):
    rex = model_factory(model_klass=EqualGainModel, randomized_value=1)

    actor_1 = rex.actors['Actor-1']
    actor_2 = rex.actors['Actor-2']
    issue_1 = rex.issues['Issue-1']
    issue_2 = rex.issues['Issue-2']

    rex.calc_nbs()
    rex.determine_positions()
    rex.calc_combinations()
    rex.determine_groups_and_calculate_exchanges()

    nbs_issue_1 = rex.nbs[issue_1]
    nbs_issue_2 = rex.nbs[issue_2]

    assert rex.nbs[issue_1] == pytest.approx(Decimal(85 + 5 / 7))
    assert rex.nbs[issue_2] == pytest.approx(Decimal(35 + 5 / 7))
    exchange = rex.exchanges[0]

    a2 = exchange.i  # type: EqualGainExchangeActor
    a1 = exchange.j  # type: EqualGainExchangeActor

    assert a2.actor == actor_2
    assert a1.actor == actor_1

    nbs_issue_1_adjusted = a1.adjust_nbs(a1.opposite_actor.demand.position)
    nbs_issue_2_adjusted = a2.adjust_nbs(a2.opposite_actor.demand.position)

    assert nbs_issue_1_adjusted == 100
    assert nbs_issue_2_adjusted == 0

    nbs_issue_1_delta = abs(nbs_issue_1 - nbs_issue_1_adjusted)
    nbs_issue_2_delta = abs(nbs_issue_2 - nbs_issue_2_adjusted)

    loss_actor_1_issue_1 = nbs_issue_1_delta * a1.supply.salience * a1.supply.power
    gain_actor_2_issue_1 = nbs_issue_1_delta * a2.demand.salience * a2.demand.power

    assert loss_actor_1_issue_1 == pytest.approx(Decimal(1000 / 7))
    assert gain_actor_2_issue_1 == pytest.approx(Decimal(6000 / 7))

    gain_actor_1_issue_2 = nbs_issue_2_delta * a1.demand.salience * a1.demand.power
    loss_actor_2_issue_2 = nbs_issue_2_delta * a2.supply.salience * a2.supply.power

    assert gain_actor_1_issue_2 == pytest.approx(Decimal(22500 / 7))
    assert loss_actor_2_issue_2 == pytest.approx(Decimal(12500 / 7))

    s = a2.supply.salience / (a1.demand.salience + a2.supply.salience)

    delta_2_u1 = loss_actor_1_issue_1 / (s * a1.demand.salience)

    assert delta_2_u1 == pytest.approx(Decimal(1000 / 225))

    delta_2_u2 = gain_actor_2_issue_1 / (s * a2.supply.salience)

    assert delta_2_u2 == pytest.approx(48)

    max_eu_actor_1 = s * a1.demand.salience * delta_2_u2 - loss_actor_1_issue_1
    assert max_eu_actor_1 == 1400

    max_eu_actor_2 = gain_actor_2_issue_1 - (s * a2.supply.salience * delta_2_u1)
    assert max_eu_actor_2 == pytest.approx(Decimal(777.7778))


def test_rex_2(model_factory):
    rex = model_factory(model_klass=EqualGainModel, randomized_value=1)

    actor_1 = rex.actors['Actor-1']
    actor_2 = rex.actors['Actor-2']
    issue_1 = rex.issues['Issue-1']
    issue_2 = rex.issues['Issue-2']

    rex.calc_nbs()
    rex.determine_positions()
    rex.calc_combinations()
    rex.determine_groups_and_calculate_exchanges()

    nbs_issue_1 = rex.nbs[issue_1]
    nbs_issue_2 = rex.nbs[issue_2]

    assert rex.nbs[issue_1] == pytest.approx(Decimal(85 + 5 / 7))
    assert rex.nbs[issue_2] == pytest.approx(Decimal(35 + 5 / 7))
    exchange = rex.exchanges[0]  # type: EqualGainExchange

    assert exchange.i.actor == actor_2
    assert exchange.j.actor == actor_1

    assert exchange.i.supply.issue == issue_2
    assert exchange.j.supply.issue == issue_1

    nbs_supply_i_adjusted = exchange.i.adjust_nbs(exchange.i.opposite_actor.demand.position)
    nbs_supply_j_adjusted = exchange.j.adjust_nbs(exchange.j.opposite_actor.demand.position)

    assert nbs_supply_i_adjusted == 0
    assert nbs_supply_j_adjusted == 100

    nbs_supply_i_delta = abs(exchange.i.nbs_0 - nbs_supply_i_adjusted)
    nbs_supply_j_delta = abs(exchange.j.nbs_0 - nbs_supply_j_adjusted)

    loss_actor_j_supply_issue = nbs_supply_j_delta * exchange.j.supply.salience * exchange.j.supply.power
    gain_actor_i_demand_issue = nbs_supply_j_delta * exchange.i.demand.salience * exchange.i.demand.power

    assert loss_actor_j_supply_issue == pytest.approx(Decimal(1000 / 7))
    assert gain_actor_i_demand_issue == pytest.approx(Decimal(6000 / 7))

    loss_actor_i_supply_issue = nbs_supply_i_delta * exchange.i.supply.salience * exchange.i.supply.power
    gain_actor_j_demand_issue = nbs_supply_i_delta * exchange.j.demand.salience * exchange.j.demand.power

    assert loss_actor_i_supply_issue == pytest.approx(Decimal(12500 / 7))
    assert gain_actor_j_demand_issue == pytest.approx(Decimal(22500 / 7))

    # We kiezen hier voor actor i
    y = gain_actor_j_demand_issue / loss_actor_i_supply_issue
    z = gain_actor_i_demand_issue / loss_actor_j_supply_issue

    assert z > y
    # if y < z: issue 2
    s = exchange.i.supply.salience / (exchange.i.supply.salience + exchange.i.opposite_actor.demand.salience)

    delta_2_uj = loss_actor_j_supply_issue / (s * exchange.i.opposite_actor.demand.salience)

    assert delta_2_uj == pytest.approx(Decimal(1000 / 225))

    delta_2_ui = gain_actor_i_demand_issue / (s * exchange.i.supply.salience)

    assert delta_2_ui == pytest.approx(48)

    max_eu_actor_j = s * exchange.j.demand.salience * delta_2_ui - loss_actor_j_supply_issue
    assert max_eu_actor_j == 1400

    max_eu_actor_i = gain_actor_i_demand_issue - (s * exchange.i.supply.salience * delta_2_uj)
    assert max_eu_actor_i == pytest.approx(Decimal(777.7778))

    exchange.calculate_maximum_utility()

    assert exchange.i.eu_max == max_eu_actor_i
    assert exchange.j.eu_max == max_eu_actor_j

    exchange.j, exchange.i = exchange.i, exchange.j

    exchange.calculate_maximum_utility()

    assert exchange.j.eu_max == max_eu_actor_i
    assert exchange.i.eu_max == max_eu_actor_j


# def test_x(model_factory):
#
#     for p in range(0, 125, 25):
#         p = decimal.Decimal(p / 100)
#
#         rex = model_factory(model_klass=EqualGainModel, randomized_value=p)
#
#         rex.calc_nbs()
#         rex.determine_positions()
#         rex.calc_combinations()
#         rex.determine_groups_and_calculate_exchanges()
#
#         exchange = rex.exchanges[0]  # type: EqualGainExchange
#
#         x = []
#
#         for _ in range(10000):
#             exchange.calculate()
#             u = random.uniform(0, 1)
#             v = random.uniform(0, 1)
#             z = decimal.Decimal(random.uniform(0, 1))
#
#             exchange.calculate_maximum_utility()
#
#             exchange.i.randomized_gain(u, v, z)
#
#             x.append(exchange.i.eu)
#         y = statistics.variance(x)
#         print(p, y)



import os

import numpy as np
import pandas as pd
from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param


def write_summary_result(conn, model_run_ids, output_directory, ai_type="before"):

    if ai_type == "before":
        x_type = "preference"
    else:
        x_type = "voting"

    df = pd.read_sql(
        """
    SELECT
      a.p as p,
      a.issue as issue,
      a.round as round,
      a.repetion as repetion,
      a.numerator / a.denominator AS mds
    FROM (SELECT
            sum(ai.position * ai.power * ai.salience) AS numerator,
            sum(ai.salience * ai.power)               AS denominator,
            r.pointer                                 AS repetion,
            i2.pointer + 1                                AS round,
            m.p,
      i.name as issue
          FROM actorissue ai
            LEFT JOIN issue i ON ai.issue_id = i.id
            LEFT JOIN actor a ON ai.actor_id = a.id
            LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
            LEFT JOIN repetition r ON i2.repetition_id = r.id
            LEFT JOIN modelrun m ON r.model_run_id = m.id        
          WHERE  ai.type = '%s' AND m.id IN (%s)
         GROUP BY m.id,r.id, i2.id, i.id) a
    """
        % (ai_type, list_to_sql_param(model_run_ids)),
        conn,
        index_col="p",
        columns=["mds"],
    )
    try:
        table_avg = pd.pivot_table(
            df,
            index=["issue", "p"],
            columns=["round"],
            values=["mds"],
            aggfunc=np.average,
        )
        table_avg.to_csv(os.path.join(output_directory, f"mds_average_{x_type}.csv"))
    except Exception as e:
        print(e)

    try:
        table_var = pd.pivot_table(
            df, index=["issue", "p"], columns=["round"], values=["mds"], aggfunc=np.var
        )
        table_var.to_csv(os.path.join(output_directory, f"mds_variance_{x_type}.csv"))
    except Exception as e:
        print(e)

    sql_2 = """SELECT issue.name, issue.id
FROM issue
INNER JOIN dataset d on issue.data_set_id = d.id
INNER JOIN modelrun m on d.id = m.data_set_id
WHERE m.id IN (%s)
GROUP BY issue.name, issue.id
ORDER BY issue.name""" % list_to_sql_param(
        model_run_ids
    )

    cursor = conn.execute_sql(sql=sql_2, params=[])
    issues = cursor.fetchall()

    # %%

    for name, issue_id in issues:
        df = pd.read_sql(
            """SELECT a.p                         as p,
       a.issue                     as issue,
       a.round                 as round,
       a.repetion                  as repetion,
       a.numerator / a.denominator AS mds
FROM (SELECT sum(ai.position * ai.power * ai.salience) AS numerator,
             sum(ai.salience * ai.power)
                                                       AS denominator,
             r.pointer
                                                       AS repetion,
             i2.pointer +1
                                                       AS round,
             m.p,
             i.name                                    as issue
      FROM actorissue ai
               LEFT JOIN issue i ON ai.issue_id = i.id
               LEFT JOIN actor a ON ai.actor_id = a.id
               LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
               LEFT JOIN repetition r ON i2.repetition_id = r.id
               LEFT JOIN modelrun m ON r.model_run_id = m.id
               LEFT JOIN dataset d ON a.data_set_id = d.id
      WHERE ai.type = '%s'
        AND m.id IN(%s)
        AND i.id = %s
      GROUP BY m.id, r.id, i2.id, i.id) a
        """
            % (ai_type, list_to_sql_param(model_run_ids), issue_id),
            conn,
            index_col="p",
            columns=["mds"],
        )

        try:
            table = pd.pivot_table(df, index=["round"], columns=["p"], values=["mds"], numeric_only=True)
            plt = table.plot()

            plt.set_title(name)
            plt.set_ylim(0, 110)

            lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)

            plt.figure.savefig(
                os.path.join(output_directory, f"mds_{name}_{x_type}.png"),
                bbox_extra_artists=(lgd,),
                bbox_inches="tight",
            )
        except Exception as e:
            print(e)


if __name__ == "__main__":
    m = Manager(os.environ.get("DATABASE_URL"))
    m.init_database()

    model_run_ids = [43, 44]

    write_summary_result(connection, model_run_ids, data_folder)



import math
from decimal import *

from decide.model.equalgain import EqualGainModel


def test_add_actor():
    model = EqualGainModel()

    a1 = model.add_actor("TestActor1")
    a2 = model.add_actor("TestActor2")

    assert len(model.actors) == 2
    assert a1 != a2


def test_first_phase(model):
    model.calc_nbs()  # tested in test_calculations.py#test_calc_nbs

    model.calc_combinations()  # tested below
    model.determine_positions()
    model.determine_groups_and_calculate_exchanges()

    totalactor_issues = 0

    for k, v in model.groups.items():
        a = len(model.groups[k]["a"])
        b = len(model.groups[k]["b"])
        c = len(model.groups[k]["c"])
        d = len(model.groups[k]["d"])

        assert sum([a, b, c, d]) == 10

        totalactor_issues += a * d
        totalactor_issues += b * c

    assert totalactor_issues == 240

    e = model.highest_gain()

    assert e.gain - Decimal(1.973684210526315789473684214) < 1e-8

    # the delta is necessary for the random component by exchanges which have an equal gain.
    # in some cases the gain of the exchanges (two exchanges with each two (unique) actors) are the same.
    # we pick random the next exchange. Theoretically it is even possible to get more exchanges with the same gain
    assert len(model.exchanges) - 239 < 2

    realized = list()
    realized.append(e)

    model.remove_invalid_exchanges(e)

    e1 = model.highest_gain()
    model.remove_invalid_exchanges(e1)
    realized.append(e1)

    assert e1.gain - Decimal(1.79263630876534102340553952) < 1e-8

    while len(model.exchanges) > 0:
        realize = model.highest_gain()
        model.remove_invalid_exchanges(realize)
        realized.append(realize)

    # assert len(realized) == 39


def test_calc_combinations(model):
    model.calc_combinations()
    combinations = len(list(model.issue_combinations))

    n = len(model.issues)
    k = 2

    assert combinations == math.factorial(n) / (math.factorial(k) * math.factorial(n - k))



import logging
import os

import pandas as pd
from pandas.errors import DataError

from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param, handle_data_frame

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)


def write_summary_result(conn, model_run_ids, output_directory):
    sql = """SELECT COUNT(*)   exchanges_count,
           SUM(ea.eu) utility_sum,
           AVG(ea.eu)        utility_avg,
           a.name as         actor,
           m.p
    FROM exchangeactor ea
             JOIN actor a on ea.actor_id = a.id
             JOIN exchange e on ea.id = e.i_id
             JOIN iteration i on e.iteration_id = i.id
             JOIN repetition r on i.repetition_id = r.id
             JOIN modelrun m on r.model_run_id = m.id
    WHERE m.id in (%s)
    GROUP BY a.id, m.p;""" % list_to_sql_param(model_run_ids)

    df = pd.read_sql_query(
        sql=sql,
        con=conn,
        index_col=['actor', 'p'],
    )

    # %%

    file_name = os.path.join(output_directory, 'exchanges_count.{}')

    try:

        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['exchanges_count']),
            file_name=file_name,
            title='AVG count of the executed exchanges for a repetition'
        )
    except DataError as ex:
        logging.exception(ex)
        print(ex)
    try:
        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['utility_sum']),
            file_name=os.path.join(output_directory, 'expected_utility_sum.{}'),
            title='AVG Sum of the expected utility per repetition'
        )

        handle_data_frame(
            df=pd.pivot_table(df, index=['p'], columns=['actor'], values=['utility_avg']),
            file_name=os.path.join(output_directory, 'expected_utility_avg.{}'),
            title='Average of the expected utility'
        )
    except DataError as ex:
        logging.exception(ex)
        print(ex)


if __name__ == '__main__':
    m = Manager(os.environ.get('DATABASE_URL'))
    m.init_database()

    model_run_id = 1

    write_summary_result(connection, [model_run_id], data_folder)



from unittest import TestCase

from decide.model import base


class TestActorIssue(TestCase):
    def test_determinePosition(self):
        position = 50
        salience = 0.25
        power = 0.10

        actor = base.Actor("Test")
        issue = base.Issue("Test")

        actorIssue = base.ActorIssue(
            actor=actor, issue=issue, position=position, power=power, salience=salience
        )

        self.assertEqual(actorIssue.power, power)
        self.assertEqual(actorIssue.salience, salience)
        self.assertEqual(actorIssue.position, position)

        self.assertTrue(actorIssue.is_left_to_nbs(51), position)



from typing import List

from PyQt5 import QtWidgets, QtCore
from decide.data import types
from decide.qt import utils


class InputFilGrid:
    pass


class DynamicFormLayout(QtWidgets.QGridLayout):
    def __init__(self, main_window, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.main_window = main_window
        self.checkboxes = []
        self.objects = []
        self.row_pointer = 0

        self.setAlignment(QtCore.Qt.AlignTop)

    def clear(self):
        """
        Set the parent of all child widgets to None so they will be deleted
        """
        for i in reversed(range(self.count())):
            self.itemAt(i).widget().setParent(None)

    def add_row(self, *widgets):

        for index, widget in enumerate(widgets):
            self.addWidget(widget, self.row_pointer, index)

        self.row_pointer += 1

    def set_values(self, objects):
        self.objects = objects
        self.objects.sort()
        self.clear()

    def get_selected(self):

        selected = []

        for checkbox in self.checkboxes:  # type: QtWidgets.QCheckBox
            if checkbox.isChecked():
                selected.append(checkbox.objectName())

        return selected


class IssueWidget(DynamicFormLayout):
    """
    Form layout for customizing the Issue selection
    """

    def set_issues(self, issues: List[types.PartialIssue]):
        self.set_values(issues)
        self.add_row(
            QtWidgets.QLabel("Issue"),
            QtWidgets.QLabel("Lower"),
            QtWidgets.QLabel("Upper"),
        )
        for issue in issues:
            checkbox = QtWidgets.QCheckBox(issue.name)

            lower = QtWidgets.QLabel(str(round(issue.lower, 2)))
            upper = QtWidgets.QLabel(str(round(issue.upper, 2)))

            checkbox.setObjectName(issue.name)
            checkbox.stateChanged.connect(self.state_changed)

            if (
                issue in self.main_window.settings.selected_issues
                or len(self.main_window.settings.selected_issues) == 0
            ):
                checkbox.setChecked(True)

            self.checkboxes.append(checkbox)

            self.add_row(checkbox, lower, upper)

    def state_changed(self):
        self.main_window.overview_widget.update_widget()

    def save(self):
        self.main_window.settings.selected_issues = self.get_selected()


class ActorWidget(DynamicFormLayout):
    """
    Form layout for customizing the Actor selection
    """

    def set_actors(self, actors: List[types.PartialActor]):
        self.set_values(actors)

        for actor in actors:
            checkbox = QtWidgets.QCheckBox(str(actor.fullname))
            checkbox.setObjectName(actor.id)

            if (
                actor in self.main_window.settings.selected_actors
                or len(self.main_window.settings.selected_actors) == 0
            ):
                checkbox.setChecked(True)

            checkbox.stateChanged.connect(self.state_changed)

            self.checkboxes.append(checkbox)
            self.add_row(checkbox)

    def state_changed(self):
        self.main_window.overview_widget.update_widget()

    def save(self):
        self.main_window.settings.selected_actors = self.get_selected()


class ActorIssueWidget(DynamicFormLayout):
    def __init__(self, *args, **kwargs):
        super(ActorIssueWidget, self).__init__(*args, **kwargs)

    def set_actor_issues(self, actor_issues=List[types.ActorIssue]):
        self.set_values(actor_issues)

        heading = ["Issue", "Actor", "Position", "Salience", "Power"]

        self.add_row(*[QtWidgets.QLabel(x) for x in heading])

        for actor_issue in actor_issues:
            issue = QtWidgets.QLabel(str(actor_issue.issue))
            actor = QtWidgets.QLabel(str(actor_issue.actor))
            position = QtWidgets.QLabel(str(round(actor_issue.position, 2)))
            salience = QtWidgets.QLabel(str(actor_issue.salience))
            power = QtWidgets.QLabel(str(actor_issue.power))

            self.add_row(issue, actor, position, salience, power)


class SummaryWidget(DynamicFormLayout):
    def __init__(
        self, main_window, settings, data, actor_widget, issue_widget, *args, **kwargs
    ):
        """
        :type settings: decide.qt.mainwindow.settings.ProgramSettings
        :type data: decide.qt.mainwindow.gui.ProgramData
        :type actor_widget: decide.qt.mainwindow.widgets.ActorWidget
        :type issue_widget: decide.qt.mainwindow.widgets.IssueWidget
        """
        super(SummaryWidget, self).__init__(main_window, *args, **kwargs)

        self.settings = settings
        self.data = data
        self.actor_widget = actor_widget
        self.issue_widget = issue_widget
        self.run_name = QtWidgets.QLineEdit()
        self.run_name.setText(self.settings.run_name)
        self.run_name.textChanged.connect(self.update_run_name)

    def update_run_name(self, value):
        self.settings.run_name = value

    def update_widget(self):

        self.clear()

        self.add_text_row("Input", self.settings.input_filename, self.test_callback)

        if (
            self.settings.output_directory is None
            or self.settings.output_directory == "None"
        ):
            self.add_text_row(
                "Output directory",
                "set output directory",
                self.main_window.select_output_dir,
            )
        else:
            self.add_text_row(
                "Output directory", self.settings.output_directory, self.test_callback
            )

        self.add_row(QtWidgets.QLabel("Run name"), self.run_name)

        self.add_text_row(
            "", "You can make use of the variable {rounds} and {repetitions} in run name"
        )

        settings = self.main_window.settings  # type: ProgramSettings

        self.add_text_row("p-values", ", ".join(settings.model_variations))

        output_selection = []

        if settings.issue_development_csv:
            output_selection.append("Issue development values [.csv]")

        if settings.summary_only:
            output_selection.append("Summary only [.csv]")

        if settings.output_sqlite:
            output_selection.append("Sqlite [database]")

        if settings.exchanges_csv:
            output_selection.append("Exchange values [.csv]")

        if settings.externalities_csv:
            output_selection.append("Externalities [.csv]")

        if settings.voting_positions:
            output_selection.append("Show Voting positions [.csv]")

        self.main_window.set_start_button_state()

    def add_text_row(self, label, value, callback=None):

        if callback:

            if len(value) > 50:
                value_x = "..." + value[-50:]
            else:
                value_x = value

            value_label = QtWidgets.QLabel(
                '<a href="{}">{}</a>'.format(value, value_x)
            )
            value_label.linkActivated.connect(callback)
        else:
            value_label = QtWidgets.QLabel(value)

        self.add_row(QtWidgets.QLabel(label), value_label)

    def test_callback(self, link):
        utils.open_file_natively(link)


class MenuBar(QtWidgets.QMenuBar):
    def __init__(self, main_window, *args, **kwargs):
        """
        :type main_window: decide.qt.mainwindow.gui.DecideMainWindow
        """
        super().__init__(*args, **kwargs)

        self.main_window = main_window
        self.settings = main_window.settings

        file_menu = self.addMenu("&File")
        # output_menu = self.addMenu("&Output")

        self.output_sqlite = QtWidgets.QAction("&sqlite", self)
        self.output_sqlite.setCheckable(True)
        self.output_sqlite.setChecked(True)

        self.issue_development_csv = QtWidgets.QAction("Issue development (.csv)", self)
        self.issue_development_csv.setCheckable(True)
        self.issue_development_csv.setChecked(True)

        self.externalities_csv = QtWidgets.QAction("Externalities (.csv)", self)
        self.externalities_csv.setCheckable(True)
        self.externalities_csv.setChecked(True)

        self.exchanges_csv = QtWidgets.QAction("Exchanges (.csv)", self)
        self.exchanges_csv.setCheckable(True)
        self.exchanges_csv.setChecked(True)

        self.voting_positions = QtWidgets.QAction("Show voting positions", self)
        self.voting_positions.setCheckable(True)
        self.voting_positions.setChecked(True)

        self.summary_only = QtWidgets.QAction("Summary only (.csv)", self)
        self.summary_only.setCheckable(True)
        self.summary_only.setChecked(True)

        self.open_data_view = QtWidgets.QAction("New data file", self)
        self.open_data_view.triggered.connect(self.main_window.open_data_view)

        open_action = QtWidgets.QAction("&Open data file", self)
        open_action.triggered.connect(self.main_window.open_input_data)

        edit_action = QtWidgets.QAction("&Edit data file", self)
        edit_action.triggered.connect(
            self.main_window.open_current_input_window_with_current_data
        )

        save_settings = QtWidgets.QAction("&Save settings", self)
        save_settings.triggered.connect(self.main_window.save_settings)

        output_dir_action = QtWidgets.QAction("&Set output directory", self)
        output_dir_action.triggered.connect(self.main_window.select_output_dir)

        # debug = self.addMenu("Debug")
        log_action = QtWidgets.QAction("Show log window", self)
        log_action.triggered.connect(self.main_window.show_debug_dialog)

        file_menu.addAction(open_action)
        self.recently_opened_menu(file_menu)
        file_menu.addAction(self.open_data_view)
        file_menu.addSeparator()
        file_menu.addAction(edit_action)
        file_menu.addSeparator()
        file_menu.addAction(save_settings)
        file_menu.addAction(output_dir_action)
        file_menu.addSeparator()
        file_menu.addAction(log_action)
        # debug.addAction(log_action)

        # error_report = QtWidgets.QAction("Send error report", self)
        # error_report.triggered.connect(self.main_window.show_error_report_dialog)
        # debug.addAction(error_report)

        # error_report_2 = QtWidgets.QAction("Trigger error", self)
        # error_report_2.triggered.connect(self.trigger_error)
        # debug.addAction(error_report_2)

    def recently_opened_menu(self, menu: QtWidgets.QMenu):
        sub_menu = QtWidgets.QMenu("Recently opened", menu)
        menu.addMenu(sub_menu)

        for item in self.settings.recently_opened:
            action = QtWidgets.QAction(item, self)
            action.triggered.connect(self.open_input_data)
            sub_menu.addAction(action)

    def open_input_data(self):
        sender = self.sender()  # type: QtWidgets.QAction

        self.main_window.init_ui_data(sender.text())

    def trigger_error(self):
        raise Exception("Error triggered")

    def load(self):
        """
        Copy the values from the ProgramSettings object
        """

        settings = self.settings.__dict__.items()

        for key, value in settings:
            if hasattr(self, key):
                attr = getattr(self, key)

                if isinstance(value, bool):  # type: QtWidgets.QAction
                    attr.setChecked(value)
                else:
                    attr.setValue(value)

    def save(self):
        """
        Set the attributes to the ProgramSettings object
        """
        settings = self.__dict__.items()

        for key, value in settings:
            if hasattr(self.settings, key):
                attr = getattr(self, key)

                if isinstance(attr, QtWidgets.QAction):
                    setattr(self.settings, key, value.isChecked())
                else:
                    setattr(self.settings, key, value.value())



import logging


class ModelLoop(object):
    """
    Helps performing all the actions in the correct order
    """

    def __init__(self, model, event_handler: "observer.Observable", repetition: int):
        self.model = model
        self.event_handler = event_handler
        self.iteration_number = 0
        self.repetition_number = repetition

    def loop(self):
        self.model.calc_nbs()
        self.model.determine_positions()
        self.model.calc_combinations()
        self.model.determine_groups_and_calculate_exchanges()

        realized = []

        # call the event for beginning the loop
        self.event_handler.before_loop(self.iteration_number, self.repetition_number)

        while len(self.model.exchanges) > 0:

            realize_exchange = self.model.highest_gain()  # type: base.AbstractExchange

            if realize_exchange and realize_exchange.is_valid:
                removed_exchanges = self.model.remove_invalid_exchanges(
                    realize_exchange
                )

                realized.append(realize_exchange)

                self.event_handler.removed_exchanges(removed_exchanges)
                self.event_handler.execute_exchange(realize_exchange)
            else:
                if self.model.VERBOSE:
                    logging.info(realize_exchange)

        # call the event for ending the loop
        self.event_handler.after_loop(
            realized=realized,
            iteration=self.iteration_number,
            repetition=self.repetition_number,
        )

        for exchange in realized:
            self.model.actor_issues[exchange.i.supply.issue][
                exchange.i.actor
            ].position = exchange.i.y
            self.model.actor_issues[exchange.j.supply.issue][
                exchange.j.actor
            ].position = exchange.j.y

        # calc the new MDS on the voting positions and fire the event for ending this loop
        self.model.calc_nbs()
        self.event_handler.end_loop(
            iteration=self.iteration_number, repetition=self.repetition_number
        )

        # calculate for each realized exchange there new start positions
        for exchange in realized:
            pi = exchange.i.new_start_position()
            self.model.actor_issues[exchange.i.supply.issue][
                exchange.i.actor
            ].position = pi

            pj = exchange.j.new_start_position()
            self.model.actor_issues[exchange.j.supply.issue][
                exchange.j.actor
            ].position = pj

        self.iteration_number += 1



import datetime
from collections import defaultdict
from typing import List

from decide import results
from decide.data import database as db
from decide.data.database import connection
from decide.model import calculations
from decide.model.base import AbstractExchange, Actor, Issue, AbstractExchangeActor
from decide.model.observers.observer import Observer, Observable


class SQLiteObserver(Observer):
    """
    Observer to store all the data in a sqlite database
    """

    def __init__(self, observable: "Observable", output_directory: str):
        super().__init__(observable)

        self.repetitions = {}
        self.iterations = defaultdict(lambda: {})

        self.actors = {}
        self.issues = {}
        self.model_run_ids = []
        self.data_set = None
        self.model_run = None

        if not output_directory.endswith(".db") and output_directory != ":memory:":
            output_directory += "/decide-data.sqlite.db"
            self.log("logging to database {}".format(output_directory))

        if not output_directory.startswith('sqlite:///'):
            output_directory = "sqlite:///" + output_directory

        self.database_path = output_directory

    def before_model(self):

        # initialize the database
        manager = db.Manager(self.database_path)
        manager.init_database()
        manager.create_tables()

        with db.connection.atomic():

            data_set, created = db.DataSet.get_or_create(
                name=self.model_ref.data_set_name
            )

            self.data_set = data_set

            for actor in self.model_ref.actors.values():  # type: Actor
                actor, created = db.Actor.get_or_create(
                    name=actor.name, key=actor.actor_id, data_set=self.data_set
                )
                self.actors[actor] = actor

            for issue in self.model_ref.issues.values():  # type: Issue
                issue, created = db.Issue.get_or_create(
                    name=issue.name,
                    key=issue.issue_id,
                    lower=issue.lower,
                    upper=issue.upper,
                    data_set=self.data_set,
                )
                self.issues[issue] = issue

    def before_repetitions(self, repetitions, iterations, randomized_value=None):
        """
        Create a new data set when needed and add all the actors

        """
        # setup
        self.repetitions = {}
        self.iterations = defaultdict(lambda: {})

        # create a data set row or find existing one
        # add the Issues and Actors when they are not present
        with db.connection.atomic():
            self.model_run = db.ModelRun.create(
                p=randomized_value or self.model_ref.randomized_value,
                iterations=iterations,
                repetitions=repetitions,
                data_set=self.data_set,
            )

    def before_iterations(self, repetition):
        with db.connection.atomic():
            repetition = db.Repetition.create(
                pointer=repetition,
                model_run=self.model_run,
                p=self.model_ref.randomized_value,
            )

            self.repetitions[repetition] = repetition

    def before_loop(self, iteration: int, repetition: int):
        with db.connection.atomic():
            self._write_actor_issues(iteration, repetition)

    def after_loop(
            self, realized: List[AbstractExchange], iteration: int, repetition: int
    ):
        iteration = self.iterations[repetition][iteration]

        with db.connection.atomic():
            for exchange in realized:
                db_exchange = db.Exchange()
                db_exchange.i = self._create_exchange_actor(exchange.i)
                db_exchange.j = self._create_exchange_actor(exchange.j)

                if db_exchange.i.eu < 1e-10:
                    pass

                if db_exchange.j.eu < 1e-10:
                    pass

                db_exchange.iteration = iteration
                db_exchange.save()

                self._write_externalities(exchange, db_exchange)

    def end_loop(self, iteration: int, repetition: int):
        with db.connection.atomic():
            self._write_actor_issues(iteration, repetition, "after")

    def after_repetitions(self):

        self.model_run.finished_at = datetime.datetime.now()
        self.model_run.save()
        self.model_run_ids.append(self.model_run.id)

        results.covariance.write_result(connection, self.model_run.iterations - 1, self.model_run.id, self.output_directory)

    def after_model(self):
        try:
            results.externalities.write_summary_result(db.connection, self.model_run_ids, self.output_directory)
            results.descriptives.write_summary_result(db.connection, self.model_run_ids, self.output_directory)
            results.issuecomparison.write_summary_result(db.connection, self.model_run_ids, self.output_directory)
            results.nashbargainingsolution.write_summary_result(db.connection, self.model_run_ids, self.output_directory)
            results.nashbargainingsolution.write_summary_result(db.connection, self.model_run_ids, self.output_directory, "after")
        except Exception as e:
            print(e)

    def _write_externalities(
        self, exchange: AbstractExchange, db_exchange: db.Exchange
    ):

        issue_set_key = self.model_ref.create_existing_issue_set_key(
            exchange.p, exchange.q
        )
        inner = exchange.get_inner_groups()

        for actor in self.actors:

            externality = db.Externality()
            externality.actor = actor
            externality.exchange = db_exchange
            externality.supply = db_exchange.i.supply_issue
            externality.demand = db_exchange.i.demand_issue
            externality.iteration = db_exchange.iteration

            externality_size = calculations.actor_externalities(
                actor, self.model_ref, exchange
            )

            is_inner = self.model_ref.is_inner_group_member(
                str(actor.key), inner, issue_set_key
            )

            if actor.key == exchange.i.actor.actor_id:
                externality.own = exchange.i.eu
            elif actor.key == exchange.j.actor.actor_id:
                exchange.own = exchange.j.eu
            else:
                if externality_size < 0:
                    if is_inner:
                        externality.inner_negative = externality_size
                    else:
                        externality.outer_negative = externality_size
                else:
                    if is_inner:
                        externality.inner_positive = externality_size
                    else:
                        externality.outer_positive = externality_size

            externality.save()

    def _write_actor_issues(self, iteration: int, repetition: int, _type="before"):

        with db.connection.atomic():

            repetition = self.repetitions[repetition]
            iteration, _ = db.Iteration.get_or_create(
                pointer=iteration, repetition=repetition
            )

            self.iterations[repetition][iteration] = iteration

            for (
                issue_obj,
                actors,
            ) in self.model_ref.actor_issues.items():
                for actor_obj, actor_issue in actors.items():
                    db.ActorIssue.create(
                        issue=self.issues[issue_obj.issue_id],
                        actor=self.actors[actor_obj.actor_id],
                        power=actor_issue.power,
                        salience=actor_issue.salience,
                        position=actor_issue.position,
                        iteration=iteration,
                        type=_type,
                    )

    def _create_exchange_actor(self, i: AbstractExchangeActor):

        exchange_actor = db.ExchangeActor()
        exchange_actor.actor = self.actors[i.actor]
        exchange_actor.supply_issue = self.issues[i.supply.issue]
        exchange_actor.demand_issue = self.issues[i.demand.issue]
        exchange_actor.eu = i.eu
        exchange_actor.x = i.supply.position
        exchange_actor.y = i.y
        exchange_actor.demand_position = i.opposite_actor.demand.position
        exchange_actor.save()

        return exchange_actor



import collections
import csv
import os
from collections import defaultdict

from .. import base
from .. import calculations
from ..observers import observer


class Externalities(observer.Observer):
    """
    There are three stages of externalities: by exchange, by issue combination and by actor. This class writes at the 
    end of each loop the externalities to the filesystem. 
    """

    def __init__(self, observable: observer.Observable, summary_only=False):
        super().__init__(observable)
        self.actor_totals = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        self.summary_only = summary_only

    def _setup(self):
        """
        Initializes the class attributes for reusable purpose        
        """
        self.connections = {}
        self.actors = defaultdict(lambda: defaultdict(int))
        self.exchanges = []

        self.database_objects = []

    def _add_exchange_set(self, externalities, realized, model, inner, issue_set_key):
        """
        Checks if an externality is own, inner or outer and positive or negative. This method also updates the actor
        externalities but returns the exchange as an set.         
        """

        exchange_set = defaultdict(int)

        for actor, value in externalities.items():

            if (
                actor == realized.i.actor or actor == realized.j.actor
            ):  # own, always positive
                key = "own"
                value = realized.gain  # TODO hotfix, should not be needed.
            else:
                is_inner = self.model_ref.is_inner_group_member(
                    actor, inner, issue_set_key
                )

                if value > 0:  # positive
                    if is_inner:  # inner
                        key = "ip"
                    else:  # outer
                        key = "op"
                else:  # negative
                    if is_inner:  # inner
                        key = "in"
                    else:  # outer
                        key = "on"

            self.actors[actor][key] += value
            exchange_set[key] += value

        return exchange_set

    def _add_or_update_issue_set(self, issue_set_key, realized, exchange_set):
        """
        Create a new entry or adds the value to the existing connection
        :param issue_set_key: 
        :param realized: 
        :param exchange_set: 
        :return: 
        """

        if issue_set_key in self.connections:
            for key, value in exchange_set.items():
                self.connections[issue_set_key][key] += value
        else:
            self.connections[issue_set_key] = exchange_set
            self.connections[issue_set_key]["first"] = realized.p
            self.connections[issue_set_key]["second"] = realized.q
            # end if

    @staticmethod
    def _calculate_externalities(model, realized):
        """
        Loops over all the actors and calculates each actor his externalities
        :param model: 
        :param realized: 
        :return: 
        """

        results = {}

        for actor in model.actors:
            results[actor] = calculations.actor_externalities(actor, model, realized)

        return results

    def _create_directories(self, repetition):
        """
        Helper for creating the output directory        
        """
        if not os.path.exists(
            "{0}/externalities/{1}".format(self.output_directory, repetition)
        ):
            os.makedirs(
                "{0}/externalities/{1}".format(self.output_directory, repetition)
            )

    def _ordered_actors(self):
        """
        Helper for outputting the actors in a consisting order        
        """
        return collections.OrderedDict(sorted(self.actors.items())).items()

    def _sum(self, ordered_actors, iteration, repetition):
        """
        sum the dictionary keys
        """

        for key, value in ordered_actors:
            actor = self.actor_totals[key][iteration]

            actor["ip"].append(value["ip"])
            actor["in"].append(value["in"])
            actor["op"].append(value["op"])
            actor["on"].append(value["on"])
            actor["own"].append(value["own"])

    def execute_exchange(self, exchange: base.AbstractExchange):
        """
        Calculations of the externalities performed each round after an exchange is executed
        :param exchange: AbstractExchange        
        """

        issue_set_key = self.model_ref.create_existing_issue_set_key(
            exchange.p, exchange.q
        )

        inner = exchange.get_inner_groups()

        externalities = self._calculate_externalities(self.model_ref, exchange)

        exchange_set = self._add_exchange_set(
            externalities, exchange, self.model_ref, inner, issue_set_key
        )

        self._add_or_update_issue_set(issue_set_key, exchange, exchange_set)

        self.exchanges.append(
            [
                exchange.i.actor,
                exchange.i.supply.issue,
                exchange.j.actor,
                exchange.j.supply.issue,
                exchange_set["ip"],
                exchange_set["in"],
                exchange_set["op"],
                exchange_set["on"],
                exchange_set["own"],
            ]
        )

    def before_repetitions(self, repetitions, iterations, randomized_value=None):
        """
        create storage units for each repetition and each iteration
        :return:
        """
        pass

    def before_loop(self, iteration: int, repetition: int):
        self._setup()
        self._create_directories(repetition)

    def end_loop(self, iteration: int, repetition: int):
        """
        Write the data to the filesystem and append the storage for the summary
        :param repetition:
        :param iteration: int the current iteration round
        :return: 
        """
        ordered_actors = self._ordered_actors()

        self._sum(ordered_actors, iteration, repetition)

        if self.summary_only:
            return

        with open(
            "{0}/externalities/{2}/externalities.{1}.csv".format(
                self.output_directory, iteration + 1, repetition
            ),
            "w",
        ) as csv_file:
            writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")

            # headings
            writer.writerow(
                [
                    "Actor",
                    "Inner Positive",
                    "Inner Negative",
                    "Outer Positive",
                    "Outer Negative",
                    "Own",
                ]
            )

            for key, value in ordered_actors:
                writer.writerow(
                    [
                        key,
                        value["ip"],
                        value["in"],
                        value["op"],
                        value["on"],
                        value["own"],
                    ]
                )

            writer.writerow([])
            writer.writerow(["Connections"])
            writer.writerow(
                [
                    "first",
                    "second",
                    "inner pos",
                    "inner neg",
                    "sx outer pos",
                    "outer neg",
                    "own",
                ]
            )

            for key, value in self.connections.items():
                writer.writerow(
                    [
                        value["first"],
                        value["second"],
                        value["ip"],
                        value["in"],
                        value["op"],
                        value["on"],
                        value["own"],
                    ]
                )

            writer.writerow([])
            writer.writerow(["Realizations"])
            writer.writerow(
                [
                    "first",
                    "supply",
                    "second",
                    "supply ",
                    "inner pos",
                    "inner neg",
                    "outer pos",
                    "outer neg",
                    "own",
                ]
            )

            for realizations in self.exchanges:
                writer.writerow(realizations)
                #
                # if insert_db:
                #     from decide.data import database as db
                #
                #     with db.connection.atomic():
                #         db.Externality.create()

    def after_repetitions(self):
        """
        Write the summary's
        """

        if not os.path.exists(
            "{0}/externalities/summary".format(self.output_directory)
        ):
            os.makedirs("{0}/externalities/summary".format(self.output_directory))

        file = defaultdict(list)

        for actor, iterations in collections.OrderedDict(
            sorted(self.actor_totals.items())
        ).items():

            for iteration, externalities in iterations.items():
                row = [actor]

                _in = self._sum_var(externalities["ip"])
                row.append(_in[0])
                row.append(_in[1])
                _in = self._sum_var(externalities["in"])
                row.append(_in[0])
                row.append(_in[1])
                _in = self._sum_var(externalities["op"])
                row.append(_in[0])
                row.append(_in[1])
                _in = self._sum_var(externalities["on"])
                row.append(_in[0])
                row.append(_in[1])
                _in = self._sum_var(externalities["own"])
                row.append(_in[0])
                row.append(_in[1])

                file[iteration].append(row)

        for key, value in file.items():

            with open(
                "{0}/externalities/summary/{1}.csv".format(self.output_directory, key),
                "w",
            ) as csv_file:
                writer = csv.writer(csv_file, delimiter=";", lineterminator="\n")

                # headings
                writer.writerow(
                    [
                        "Actor",
                        "Inner Positive",
                        "",
                        "Inner Negative",
                        "",
                        "Outer Positive",
                        "",
                        "Outer Negative",
                        "",
                        "Own" "",
                    ]
                )

                for row in value:
                    writer.writerow(row)

    def _sum_var(self, items: list):

        if len(items) == 0:
            return 0, 0

        total = sum(items)
        avg = total / len(items)

        if total == 0:
            return 0, 0

        variance = sum([pow(x - avg, 2) for x in items]) / len(items)

        return str(avg), str(variance)



import os

import pytest
import typesystem
from decide import input_folder
from typesystem import ValidationError

from .. import reader, types


def test_cop2():
    data_file = reader.InputDataFile.open(os.path.join(input_folder, "cop21.csv"))

    assert len(data_file.errors) == 0
    assert len(data_file.rows) == 235


def test_validation_error():
    data = [
        [types.PartialActor.starts_with, "Actor", "Actor description"],
        [types.PartialIssue.starts_with, 'Issue', 'Issue description'],
        [types.IssuePosition.starts_with, 'Issue', 0, 'Issue position description'],
        [types.IssuePosition.starts_with, 'Issue', 100, 'Issue position description'],
        [types.ActorIssue.starts_with, "Actor", "Issue", 150, 0.5, 0.5, "Comment", "2", ],
    ]

    input_file = reader.InputDataFile()
    input_file.parse_rows(data)

    assert len(input_file.errors) == 0

    issue = input_file.issues['Issue']

    input_file.update_issues_with_positions()

    assert issue.lower == 0
    assert issue.upper == 100

    input_file.validate_actor_issue_positions()

    assert len(input_file.errors) == 1


def test_hash():
    issue = types.PartialIssue()
    issue.name = 'x'

    data = {issue: issue}

    assert data['x'] == issue


def test_parse_actor_issue_row():
    data = [
        types.ActorIssue.starts_with,
        "Actor",
        "Issue",
        50,
        0.5,
        0.5,
        "Comment",
        "2",
    ]
    obj = reader.csv_row_to_type(data)  # type: types.ActorIssue

    assert obj.issue == "Issue"
    assert obj.comment == "Comment 2", obj.comment


def test_parse_actor_issue_row_power_to_high():
    data = [
        types.ActorIssue.starts_with,
        "Actor",
        "Issue",
        50,
        111.5,
        100.5,
        "Comment",
        "2",
    ]

    with pytest.raises(typesystem.ValidationError):
        reader.csv_row_to_type(data)  # type: types.ActorIssue


def test_parse_actor_issue_row_power_to_low():
    data = [
        types.ActorIssue.starts_with,
        "Actor",
        "Issue",
        50,
        -1.5,
        0.5,
        "Comment",
        "2",
    ]

    with pytest.raises(typesystem.ValidationError):
        reader.csv_row_to_type(data)  # type: types.ActorIssue


def test_squash():
    data = ["x", "y", "z"]

    reader.squash(2, [])

    squash = reader.squash(2, data)

    assert squash[-1] == "y z", squash[-1]

    assert reader.squash(3, data) == data, reader.squash(3, data)


def test_issue_validate_interval():
    issue = types.PartialIssue()
    issue.lower = 0
    issue.upper = 100

    issue.validate_interval()

    with pytest.raises(ValidationError):
        issue.lower = 120
        issue.validate_interval()


def test_actor_issue_validate_position():
    issue = types.PartialIssue()
    issue.lower = 0
    issue.upper = 100

    actor_issue = types.ActorIssue()
    actor_issue.position = 5

    actor_issue.validate_position(issue)

    with pytest.raises(ValidationError):
        actor_issue.position = 999
        actor_issue.validate_position(issue)



import pytest

from decide.model.base import Actor, Issue


def test_actor_functions():
    actor1 = Actor("Test")
    actor2 = Actor("Test")
    actor3 = Actor("Test 123")

    assert actor1.name == "Test"
    assert actor1.actor_id == "Test"
    assert actor2 == "Test"
    assert actor3.name == "Test 123"
    assert actor3.actor_id == "Test 123"
    assert str(actor3) == actor3.name

    # actor comparison happens on the key attribute. So the following is correct
    assert actor1 == actor1
    assert actor1 == actor2
    assert actor1 != actor3

    assert actor1 is not None

    # self.assertFalse(actor1 is None)
    with pytest.raises(NotImplementedError):
        assert actor1 == Issue("mock")

    # testing hash index functions
    store = {actor1: actor1, actor2: actor2, actor3: actor3}

    assert store[actor1] == actor1
    assert store[actor2] == actor1
    assert store[actor3] != actor1

    assert actor1 < actor3

    with pytest.raises(ValueError):
        assert actor1 < Issue("Mock")



import logging
import sys

from ..observers.observer import Observer, Observable


class Logger(Observer):
    LOG_LEVEL = 2

    def update(self, observable, notification_type, **kwargs):
        if notification_type == Observable.LOG and Logger.LOG_LEVEL > 1:
            logging.info("LOG:", kwargs["message"])
        elif notification_type == Observable.EXECUTED and Logger.LOG_LEVEL > 2:
            logging.info("LOG: Executed exchange")
        elif notification_type == Observable.REMOVED and Logger.LOG_LEVEL > 2:
            logging.info("LOG: Removed {0} exchange(s)".format(len(kwargs["removed"])))
        elif notification_type == Observable.FINISHED_ROUND and Logger.LOG_LEVEL > 1:

            sys.stdout.write(
                "\rLOG: Finished round {0} with {1} exchanges".format(
                    kwargs["iteration"], len(kwargs["realized"])
                )
            )
        elif notification_type == Observable.CLOSE and Logger.LOG_LEVEL > 1:
            sys.stdout.write("\n")



from collections import defaultdict
from decimal import Decimal

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QListWidgetItem

from decide.data.types import ActorIssue
from decide.qt.inputwindow import signals
from decide.qt.inputwindow.models import (
    ActorInputModel,
    IssueInputModel,
    ActorIssueInputModel,
)
from decide.qt.mainwindow.helpers import DoubleInput, clear_layout, normalize


class BoxLayout(QtWidgets.QGroupBox):
    def __init__(self, title, btn=True):
        super(BoxLayout, self).__init__(title)

        self.scroll_area = QtWidgets.QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)

        self.scroll_area_widget = QtWidgets.QWidget()
        self.scroll_area.setWidget(self.scroll_area_widget)

        self.grid_layout = QtWidgets.QGridLayout(self.scroll_area_widget)
        self.grid_layout.setAlignment(QtCore.Qt.AlignTop)

        self.layout_container = QtWidgets.QVBoxLayout()
        self.layout_container.addWidget(self.scroll_area)

        if btn:
            container = QtWidgets.QHBoxLayout()

            self.add_button = QtWidgets.QPushButton("Add")

            container.addWidget(self.add_button)

            self.layout_container.addLayout(container)
            self.layout_container.addWidget(self.add_button)

        self.setLayout(self.layout_container)

        self.actor_issues = {}

        self._row_pointer = 0

    def clear(self):
        clear_layout(self.grid_layout)

        self.actor_issues = {}
        self._row_pointer = 0

    def add_heading(self):
        raise NotImplementedError

    def add_row(self, *args, pointer=None):

        if pointer:
            row = pointer
        else:
            row = self._row_pointer

        for column, arg in enumerate(args):
            if isinstance(arg, str):
                arg = QtWidgets.QLabel(arg)

            self.grid_layout.addWidget(arg, row, column)

        self._row_pointer += 1

        return row

    def delete_row_callback(self):

        sending_button = self.sender()  # type: QtWidgets.QPushButton

        row = int(sending_button.objectName().split("-")[-1])

        obj = self.actor_issues[row]

        del self.actor_issues[row]

        for column in range(self.grid_layout.count()):

            item = self.grid_layout.itemAtPosition(row, column)
            if item:
                item.widget().deleteLater()
            self.grid_layout.removeItem(item)

        return obj


class ActorWidget(BoxLayout):
    key = "actor_box"

    NEW_ACTOR_POINTER = 0

    def __init__(self):
        super(ActorWidget, self).__init__("Actors")
        self.add_heading()
        self.add_button.clicked.connect(self.add_action)

    def add_heading(self):
        self.add_row("Actor", "Power")

    def add_delete_button(self):

        delete_button = QtWidgets.QPushButton("Delete")
        delete_button.clicked.connect(self.delete_row_callback)
        delete_button.setObjectName("actor-" + str(self._row_pointer))

        return delete_button

    def add_power_input(self, actor_input, power):

        power_input = DoubleInput()
        power_input.setValue(power)
        power_input.valueChanged.connect(actor_input.set_power)

        actor_input.widgets['power'] = power_input

        return power_input

    def add_comment_button(self):

        comment_button = QtWidgets.QPushButton("Comment")
        comment_button.clicked.connect(self.comment_window)
        comment_button.setObjectName("actor-" + str(self._row_pointer))

        return comment_button

    def add_name_input(self, actor_input, name):
        name_input = QtWidgets.QLineEdit()
        name_input.setText(name)
        name_input.textChanged.connect(actor_input.set_name)

        actor_input.widgets['id'] = name_input

        return name_input

    def add_actor(self, name=None, power=Decimal(0.0)):

        if not name:
            name = self.create_actor_name()

        actor_input = ActorInputModel(name, power)
        actor_input.id = self._row_pointer

        name_input = self.add_name_input(actor_input, name)
        power_input = self.add_power_input(actor_input, power)
        delete_button = self.add_delete_button()
        comment_button = self.add_comment_button()

        self.actor_issues[actor_input.id] = actor_input

        self.add_row(name_input, power_input, comment_button, delete_button)

        return actor_input

    @staticmethod
    def create_actor_name():

        ActorWidget.NEW_ACTOR_POINTER += 1

        return "Actor-{}".format(ActorWidget.NEW_ACTOR_POINTER)

    def add_action(self):
        a = self.add_actor()

        signals.actor_created.send(a)

    def delete_row_callback(self):
        row = super().delete_row_callback()

        signals.actor_deleted.send(row)

    def comment_window(self):
        sending_button = self.sender()  # type: QtWidgets.QPushButton

        row = int(sending_button.objectName().split("-")[-1])

        obj = self.actor_issues[row]

        text, ok_pressed = QtWidgets.QInputDialog.getMultiLineText(
            self, "Comment", "Comment", obj.comment
        )

        if ok_pressed and text != "":
            obj.comment = text


class IssueWidget(BoxLayout):
    key = "issue_box"

    NEW_ISSUE_POINTER = 0

    def __init__(self):
        super(IssueWidget, self).__init__("Issues")
        self.add_heading()
        self.add_button.clicked.connect(self.add_action)

    @staticmethod
    def create_issue_name():

        IssueWidget.NEW_ISSUE_POINTER += 1

        return "Issue-{}".format(IssueWidget.NEW_ISSUE_POINTER)

    def add_heading(self):
        self.add_row("Issue", "Lower", "Upper")

    def add_delete_button(self):

        delete_button = QtWidgets.QPushButton("Delete")
        delete_button.clicked.connect(self.delete_row_callback)
        delete_button.setObjectName("issue-" + str(self._row_pointer))

        return delete_button

    def add_comment_button(self):

        comment_button = QtWidgets.QPushButton("Comment")
        comment_button.clicked.connect(self.comment_window)
        comment_button.setObjectName("issue-" + str(self._row_pointer))

        return comment_button

    def add_issue(self, name=None, lower=Decimal(0), upper=Decimal(100.0)):
        if not name:
            name = self.create_issue_name()
        issue_input = IssueInputModel(name, lower, upper)
        issue_input.id = self._row_pointer

        lower_input = DoubleInput()
        lower_input.setValue(lower)

        upper_input = DoubleInput()
        upper_input.setValue(upper)

        name_input = QtWidgets.QLineEdit()
        name_input.setText(name)

        # call the setter on a change
        name_input.textChanged.connect(issue_input.set_name)
        lower_input.valueChanged.connect(issue_input.set_lower)
        upper_input.valueChanged.connect(issue_input.set_upper)

        issue_input.widgets['name'] = name_input
        issue_input.widgets['lower'] = lower_input
        issue_input.widgets['upper'] = upper_input

        comment_button = self.add_comment_button()
        delete_button = self.add_delete_button()

        self.actor_issues[issue_input.id] = issue_input

        self.add_row(
            name_input, lower_input, upper_input, comment_button, delete_button
        )

        return issue_input

    def add_action(self):
        issue = self.add_issue()

        signals.issue_created.send(issue)

    def delete_row_callback(self):
        row = super().delete_row_callback()

        signals.issue_deleted.send(row)

    def comment_window(self):
        sending_button = self.sender()  # type: QtWidgets.QPushButton

        row = int(sending_button.objectName().split("-")[-1])

        obj = self.actor_issues[row]

        text, ok_pressed = QtWidgets.QInputDialog.getMultiLineText(
            self, "Comment", "Comment", obj.comment
        )

        if ok_pressed and text != "":
            obj.comment = text


class ActorIssueWidget(BoxLayout):
    """
    The ActorIssueBox is an hidden box behind the scenes, so there is a single point of truth
    """

    def add_heading(self):
        pass

    def __init__(self, actor_box: ActorWidget, issue_box: IssueWidget):
        super(ActorIssueWidget, self).__init__("Hidden Box")

        self.actor_box = actor_box
        self.issue_box = issue_box

        self.actor_issues = defaultdict(lambda: dict())

        self.actors = set()
        self.issues = set()

    def clear(self):
        super(ActorIssueWidget, self).clear()
        self.actor_issues = defaultdict(lambda: dict())

        self.actors = set()
        self.issues = set()

    def delete_actor(self, row):
        if row in self.actors:
            self.actors.remove(row)

            for issue in self.issue_box.actor_issues.values():
                if row.id in self.actor_issues and issue.id in self.actor_issues[row.id]:
                    del self.actor_issues[row.id][issue.id]

    def delete_issue(self, row):
        if row in self.issues:
            self.issues.remove(row)
            for actor in self.actor_box.actor_issues.values():
                del self.actor_issues[actor.id][row.id]

    def add_issue(self, value):
        self.issues.add(value)
        for actor in self.actor_box.actor_issues.values():
            self.add_actor_issue(actor, value)

    def add_actor(self, value):
        self.actors.add(value)
        for issue in self.issue_box.actor_issues.values():
            self.add_actor_issue(value, issue)

    def add_actor_issue(
            self,
            actor: ActorInputModel,
            issue: IssueInputModel,
            actor_issue: ActorIssue = None,
    ):

        actor_issue_input = ActorIssueInputModel(actor, issue)
        actor_issue_input.id = self._row_pointer

        self.actor_issues[actor.id][issue.id] = actor_issue_input

        if actor_issue:
            actor_issue_input.set_position(actor_issue.position, silence=True)
            actor_issue_input.set_salience(actor_issue.salience, silence=True)
            actor_issue_input.set_power(actor_issue.power, silence=True)

        self._row_pointer += 1


class PositionSalienceWidget(QtWidgets.QWidget):
    def __init__(self, actor_issue_box: ActorIssueWidget, *args, **kwargs):
        super(PositionSalienceWidget, self).__init__(*args, **kwargs)

        self.actor_issue_box = actor_issue_box

        self._row_pointer = 0

        self.choices_list_widget = QtWidgets.QListWidget(self)
        self.choices_list_widget.setMaximumHeight(100)
        self.choices_list_widget.setSelectionMode(QtWidgets.QListWidget.MultiSelection)
        self.choices_list_widget.itemSelectionChanged.connect(self.redraw)

        self.choices_list_widget_items = {}

        self.container = QtWidgets.QVBoxLayout(self)
        self.container.addWidget(self.choices_list_widget)

        self.scroll_area = QtWidgets.QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)

        self.scroll_area_widget = QtWidgets.QWidget()
        self.scroll_area.setWidget(self.scroll_area_widget)

        self.grid_layout = QtWidgets.QGridLayout(self.scroll_area_widget)
        self.grid_layout.setAlignment(QtCore.Qt.AlignTop)

        self.container.addWidget(self.scroll_area)

        self.add_heading()

    def clear(self):

        self._row_pointer = 0
        clear_layout(self.grid_layout)

    def add_heading(self):
        raise NotImplementedError

    def add_row(self, *args, pointer=None):

        if pointer:
            row = pointer
        else:
            row = self._row_pointer

        for column, arg in enumerate(args):
            if isinstance(arg, str):
                arg = QtWidgets.QLabel(arg)

            self.grid_layout.addWidget(arg, row, column)

        self._row_pointer += 1

        return row

    def redraw(self):

        clear_layout(self.grid_layout)

        self.add_heading()

        for issue in self.actor_issue_box.issues:
            for actor in self.actor_issue_box.actors:

                if (
                        actor.id in self.actor_issue_box.actor_issues
                        and issue.id in self.actor_issue_box.actor_issues[actor.id]
                ):
                    actor_issue = self.actor_issue_box.actor_issues[actor.id][
                        issue.id
                    ]  # type: ActorIssueInputModel

                    self.draw_actor_issue(actor_issue)

    def draw_actor_issue(self, actor_issue: ActorIssueInputModel):
        raise NotImplementedError

    def add_choice(self, key, value):

        item = QListWidgetItem(self.choices_list_widget)
        item.setText(value)

        self.choices_list_widget_items[key] = item

        self.choices_list_widget.addItem(item)

        self.choices_list_widget.sortItems()

    def delete_choice(self, key):

        item = self.choices_list_widget_items[key]  # type: QtWidgets.QListWidgetItem

        self.choices_list_widget.takeItem(self.choices_list_widget.row(item))

    def update_choice(self, key, value):

        item = self.choices_list_widget_items[key]

        item.setText(value)

    def set_choices(self, choices):

        self.choices_list_widget.clear()

        for item in choices:
            if item.name != "":
                item_widget = QtWidgets.QListWidgetItem(self.choices_list_widget)
                item_widget.setText(item.name)
                self.choices_list_widget_items[item.id] = item_widget
                self.choices_list_widget.addItem(item_widget)

        self.choices_list_widget.sortItems()

    def is_selected(self, value):

        items = self.choices_list_widget.selectedItems()

        if len(items) == 0:
            return True

        for item in items:
            if item.text() == value:
                return True

        return False


class PositionWidget(PositionSalienceWidget):
    def add_heading(self):
        self.add_row("Issue", "Actor", "Power", "Position")

    def draw_actor_issue(self, actor_issue: ActorIssueInputModel):
        if self.is_selected(actor_issue.issue.name):
            position = DoubleInput()
            position.setValue(actor_issue.position)
            position.valueChanged.connect(actor_issue.set_position)

            actor_issue.widgets['position'] = position

            self.add_row(
                QtWidgets.QLabel(actor_issue.issue.name),
                QtWidgets.QLabel(actor_issue.actor.name),
                QtWidgets.QLabel(normalize(actor_issue.actor.power)),
                position,
            )


class SalienceWidget(PositionSalienceWidget):
    def add_heading(self):
        self.add_row("Actor", "Issue", "Power", "Position", "Salience")

    def draw_actor_issue(self, actor_issue: ActorIssueInputModel):
        if self.is_selected(actor_issue.actor.name):
            salience = DoubleInput()
            salience.setValue(actor_issue.salience)
            salience.valueChanged.connect(actor_issue.set_salience)

            self.add_row(
                QtWidgets.QLabel(actor_issue.actor.name),
                QtWidgets.QLabel(actor_issue.issue.name),
                QtWidgets.QLabel(normalize(actor_issue.actor.power)),
                QtWidgets.QLabel(normalize(actor_issue.position)),
                salience,
            )



import logging
import os
import sys
from typing import List

from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QDialog
from decide import log_filename, input_folder
from decide.data.reader import InputDataFile
from decide.qt.utils import exception_hook


class ErrorGrid(QDialog):
    """
    Visualisation of the input file in a dialog to inform the user of the errors a input file has
    """
    def __init__(self, data_file: InputDataFile, *args, **kwargs):
        super(ErrorGrid, self).__init__(*args, **kwargs)

        self.data_file = data_file
        self.row_pointer = 0

        self.main = QtWidgets.QGridLayout()

        self.init_window()

    def init_window(self):

        central_widget = QtWidgets.QWidget()
        central_widget.setLayout(self.main)

        layout = QtWidgets.QHBoxLayout(self)
        scroll_area = QtWidgets.QScrollArea(self)
        scroll_area.setWidgetResizable(True)
        scroll_area_widget_contents = QtWidgets.QWidget()
        self.main = QtWidgets.QGridLayout(scroll_area_widget_contents)
        scroll_area.setWidget(scroll_area_widget_contents)
        layout.addWidget(scroll_area)

        self.setLayout(layout)
        self.showMaximized()
        # self.setGeometry(300, 300, 400, 400)
        self.setWindowTitle("Error reporting tool")

        self.init_grid()

        self.show()

    def init_grid(self):

        for row, columns in self.data_file.rows.items():
            self.add_row(row, columns)

    def add_row(self, row, columns: List[str]):
        for column, content in enumerate(columns):

            label = QtWidgets.QLabel(str(content))

            if row in self.data_file.errors:
                label.setStyleSheet("color: red")
                error = self.data_file.errors[row]
                label.setToolTip(str(error))

            self.main.addWidget(label, row, column)


def main():

    logging.basicConfig(
        filename=log_filename,
        filemode="w",
        level=logging.DEBUG,
        format=" %(asctime)s - %(levelname)s - %(message)s",
    )

    sys.excepthook = exception_hook

    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)

    data_file = InputDataFile.open(
        os.path.join(input_folder, "kopenhagen_with_errors.csv")
    )

    error_dialog = ErrorGrid(data_file)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
